#  ------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License (MIT). See LICENSE in the repo root for license information.
#  ------------------------------------------------------------------------------------------

import functools
import logging
import multiprocessing
import shutil
import sys
import tempfile
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple, Union

import torchvision
import torch
import numpy as np
import pandas as pd
from PIL import Image, ImageDraw, ImageFont
from tqdm import tqdm
from monai.data.image_reader import WSIReader

from health_azure.argparsing import apply_overrides, parse_arguments
from health_cpath.preprocessing.loading import WSIBackend
from health_cpath.utils.montage_config import MontageConfig, create_montage_argparser
from health_cpath.utils.naming import SlideKey
from health_cpath.datasets.base_dataset import DEFAULT_DATASET_CSV, SlidesDataset
from health_ml.utils.type_annotations import TupleInt3


MONTAGE_FILE = "montage.png"
DatasetOrDataframe = Union[SlidesDataset, pd.DataFrame]
DatasetRecord = Dict[SlideKey, Any]

logger = logging.getLogger(__name__)


def add_text(image: Image, text: str, y: float = 0.9, color: TupleInt3 = (27, 77, 40), fontsize_step: int = 2) -> None:
    """Add text to a PIL image.

    :param image: Image object to which text needs to be added.
    :param text: The text that needs to be added.
    :param y: Float between 0-1 specifying the vertical position of the text (default=0.9).
    :param color: A 3-tuple indicating the fill color of the text (default = (27, 77, 40)).
    :param fontsize_step: Steps of font size to reduce if the text size is more than image size (default=2).
    """
    # This font is usually found in a path like /usr/share/fonts/truetype/dejavu
    font_path = Path('DejaVuSans.ttf')
    fontsize = 48
    draw = ImageDraw.Draw(image)
    image_size_x, image_size_y = image.size
    font = ImageFont.truetype(str(font_path), fontsize)
    _, _, text_size_x, text_size_y = draw.textbbox((0, 0), text, font)
    while text_size_x >= image_size_x:
        fontsize -= fontsize_step
        font = ImageFont.truetype(str(font_path), fontsize)
        _, _, text_size_x, text_size_y = draw.textbbox((0, 0), text, font)
    start_x = image_size_x // 2 - text_size_x // 2
    start_y = image_size_y * y - text_size_y // 2
    xy = start_x, start_y
    draw.text(xy, text, fill=color, font=font, align='center')


def load_slide_as_pil(reader: WSIReader, slide_file: Path, level: int = 0) -> Image:
    """Load a WSI as a PIL image.

    :param reader: The WSI reader for loading the slide.
    :param slide_file: The file to read from.
    :param level: Resolution downsampling level (default=0).
    :return: PIL image object corresponding to the WSI image.
    """
    image = reader.read(slide_file)
    try:
        image_array, _ = reader.get_data(image, level=level)
    except ValueError:
        logger.warning(f"Level {level} not available for {slide_file}, using level 0 instead.")
        image_array, _ = reader.get_data(image, level=0)
    array = image_array.numpy().transpose(1, 2, 0)
    to_pil = torchvision.transforms.ToPILImage()
    array_pil = to_pil(array)
    return array_pil


def _make_thumbnail(
    sample: DatasetRecord,
    reader: WSIReader,
    level: int,
    slide_size: Tuple[int, int],
    images_dir: Path,
    masks_dir: Optional[Path] = None,
    image_suffix: str = '.png',
    default_mask_color: TupleInt3 = (119, 161, 120),
) -> None:
    """Make thumbnails of the slides in slides dataset.

    :param sample: The slide dataset object dictionary for which thumbnail needs to be created.
    :param reader: The WSI reader for loading the slide.
    :param slide_size: The tuple of slide size (width, height).
    :param images_dir: The path to the `images` directory where WSI thumbnails will be stored.
    :param level: Resolution downsampling level.
    :param masks_dir: Optional path to `masks` directory where mask thumbnails will be stored.
        If `None` (default), masks thumbnails will not be created.
    :param image_suffix: Suffix of image thumbnails (default=`.png`).
    :param default_mask_color: Color of the masks (default = (119, 161, 120)).
    """
    try:
        image_pil = load_slide_as_pil(reader, sample[SlideKey.IMAGE], level)
        image_pil = image_pil.resize(slide_size)
        slide_id = sample.get(SlideKey.SLIDE_ID, "")
        # Slide IDs can be purely numeric, in those cases need to convert to str
        text_to_add = str(slide_id)
        if SlideKey.LABEL in sample:
            label = str(sample[SlideKey.LABEL])
            text_to_add += ": " + str(label)
        if text_to_add:
            add_text(image_pil, text_to_add)
        if masks_dir is not None and SlideKey.MASK in sample:
            masks_dir.mkdir(exist_ok=True)
            try:
                mask_pil = load_slide_as_pil(reader, sample[SlideKey.MASK], level=level)
                mask_pil = mask_pil.resize(slide_size, Image.NEAREST)
                mask_pil = Image.fromarray(np.asarray(mask_pil) * 255)  # for visualization
            except ValueError:
                mask_pil = Image.new("RGB", slide_size, default_mask_color)
            finally:
                mask_path = masks_dir / f"{slide_id}.png"
                mask_pil.save(mask_path)
        image_path = images_dir / f"{slide_id}{image_suffix}"
        image_pil.save(image_path)
    except Exception as ex:
        slide_id = sample.get(SlideKey.SLIDE_ID, "(no slide ID found)")
        logging.warning(f"Unable to process slide with ID '{slide_id}': {ex}")


def make_thumbnails(
    records: List[DatasetRecord],
    slide_size: Tuple[int, int],
    images_dir: Path,
    level: int,
    masks_dir: Optional[Path] = None,
    num_parallel: int = 0,
    image_suffix: str = '.png',
    backend: str = WSIBackend.CUCIM,
) -> None:
    """Make thumbnails of the slides in slides dataset.

    :param records: A list of dataframe records. The records must contain at least the columns `slide_id` and `image`.
    :param slide_size: The tuple of slide size (width, height).
    :param images_dir: The path to the `images` directory where WSI thumbnails will be stored.
    :param level: Resolution downsampling level.
    :param masks_dir: Optional path to `masks` directory where mask thumbnails will be stored (default=None).
    :param num_parallel: Number of parallel processes for thumbnail creation. Use 0 to disable parallel.
    :param image_suffix: Suffix of image thumbnails (default=`.png`).
    :param backend: The backend to use for reading the WSI (default=`cucim`).
    """
    images_dir.mkdir(exist_ok=True, parents=True)
    reader = WSIReader(backend=backend)
    func = functools.partial(
        _make_thumbnail,
        reader=reader,
        level=level,
        slide_size=slide_size,
        images_dir=images_dir,
        masks_dir=masks_dir,
        image_suffix=image_suffix,
    )
    if num_parallel > 0:
        pool = multiprocessing.Pool(num_parallel)
        map_func = pool.imap_unordered  # type: ignore
    else:
        map_func = map  # type: ignore
    progress = tqdm(map_func(func, records), total=len(records))
    list(progress)  # type: ignore
    if num_parallel > 0:
        pool.close()


def make_montage_from_dir(
    images_dir: Path, num_cols: int, masks_dir: Optional[Path] = None, image_suffix: str = '.png'
) -> Image:
    """Create the montage image from the thumbnails.

    :param images_dir: The path to the `images` directory where WSI thumbnails will be stored.
    :param num_cols: Number of columns in the montage.
    :param masks_dir: Optional path to `masks` directory where mask thumbnails will be stored (default=None).
    :param image_suffix: Suffix of image thumbnails (default=`.png`).
    :return: PIL image of the montage.
    """
    image_paths = sorted(images_dir.glob(f'*{image_suffix}'))
    if len(image_paths) == 0:
        raise ValueError(f"No thumbnail images found in {images_dir}")
    images_arrays = []
    for image_path in tqdm(image_paths):
        image_pil = Image.open(image_path)
        images_arrays.append(np.asarray(image_pil))
    images_array = np.asarray(images_arrays)
    if masks_dir is not None:
        mask_paths = sorted(masks_dir.glob('*.png'))
        # Don't process masks if there are no files present, even if the mask directory has been passed as an argument.
        if len(mask_paths) > 0:
            if len(mask_paths) != len(image_paths):
                raise ValueError("Number of masks is different from number of images.")
            masks_arrays = []
            for mask_path in tqdm(mask_paths):
                mask_pil = Image.open(mask_path)
                masks_arrays.append(np.asarray(mask_pil))
            masks_array = np.asarray(masks_arrays)
            images_array = np.concatenate((images_array, masks_array), axis=-2)
    images_tensor = torch.from_numpy(images_array).permute(0, 3, 1, 2)
    grid_tensor = torchvision.utils.make_grid(images_tensor, nrow=num_cols)
    grid_pil = torchvision.transforms.ToPILImage()(grid_tensor)
    return grid_pil


def dataset_from_folder(root_folder: Path, glob_pattern: str = "**/*") -> pd.DataFrame:
    """Create slides dataset all files in a folder. The function searches for all files in the `root_folder` and its
    subfolders, and creates a dataframe with the following columns: `image`: The absolute path of the file,
    column `slide_id`: Either the file name only if that is unique, or otherwise the path of the file relative to
    the `root_folder`.

    :param root_folder: A directory with (image) files.
    :param glob_pattern: The glob pattern to match the image files (default=`**/*`, using all files recursively in all
        subfolders).
    :return: Slides dataset.
    """
    if not root_folder.is_dir():
        raise ValueError(f"Root folder '{root_folder}' does not exist or is not a directory.")
    # Find all image files in the folder, exclude folders in the result
    image_paths = list(sorted(f for f in root_folder.glob(glob_pattern) if f.is_file()))
    file_names_only = [path.name for path in image_paths]
    # Check if file names alone are enough to make the dataset unique.
    if len(file_names_only) != len(set(file_names_only)):
        # There are duplicates when going only by file names. Hence, use full paths relative to the root folder.
        image_ids = [str(path.relative_to(root_folder)) for path in image_paths]
    else:
        # File names are unique. Hence, use them as slide IDs.
        image_ids = file_names_only
    # Mounted datasets can show up with '&' appearing escape. Clean up the image IDs if so.
    # We expect that the exclusion list shows '&' for the offending slides.
    escaped_amp = "%26"
    if any(escaped_amp in image_id for image_id in image_ids):
        logging.info(f"Some image IDs contain '{escaped_amp}', replacing that with '&'.")
        image_ids = [id.replace(escaped_amp, "&") for id in image_ids]
    return pd.DataFrame({SlideKey.SLIDE_ID: image_ids, SlideKey.IMAGE: map(str, image_paths)})


def restrict_dataset(dataset: pd.DataFrame, column: str, items: List[str], include: bool) -> pd.DataFrame:
    """Exclude or include slides from a dataset, based on values in a column.
    For example, to exclude slides with label `0` from the dataset, use:
    restrict_dataset(dataset, column='label', items=['0'], include=False).

    The items are matched with the column values using `isin` operator. The code also handles
    the case where the column in question is the dataset index.
    If the items in question are not present in the column, the result is an empty dataset (if include=True)
    or the original dataset (if include=False).

    :param dataset: Slides dataset.
    :param column: The name of the column on which the inclusion/exclusion name.
    :param items: The values that the column should match.
    :param include: If True, modify the dataset to only include the rows where the column matches a value in the
        item list in `items`. If False, modify the dataset to exclude those rows.
    :return: Filtered dataset.
    """
    if column in dataset:
        matching_rows = dataset[column].isin(items)
        if not include:
            matching_rows = ~matching_rows
        return dataset[matching_rows]
    elif dataset.index.name == column:
        # Drop or loc operations on an index column when the values do not exist raise an error. Hence, restrict
        # to existing values first.
        items = list(set(items).intersection(set(dataset.index)))
        if include:
            return dataset.loc[items]
        else:
            return dataset.drop(items)
    else:
        raise ValueError(f"Column {column} not found in dataset.")


def dataset_to_records(dataset: DatasetOrDataframe) -> List[DatasetRecord]:
    """Converts a SlidesDataset or a plain dataframe to a list of dictionaries.

    :param dataset: Slides dataset or a plain dataframe.
    """
    if isinstance(dataset, pd.DataFrame):
        return dataset.to_dict(orient='records')
    elif isinstance(dataset, SlidesDataset):
        # SlidesData overrides __getitem__, use that to convert to records
        return [dataset[i] for i in range(len(dataset))]
    else:
        raise ValueError(f"Can't convert {type(dataset)} to a list of records.")


def make_montage(
    records: List[DatasetRecord],
    out_path: Path,
    width: int = 60_000,
    level: int = 2,
    image_suffix: str = '.png',
    masks: bool = True,
    temp_dir: Optional[Union[Path, str]] = None,
    cleanup: bool = False,
    num_parallel: int = 0,
    backend: str = "cucim",
) -> None:
    """Make the montage of WSI thumbnails from a slides dataset.

    :param records: A list of dataframe records. The records must contain at least the columns `slide_id` and `image`.
    :param out_path: The output path where the montage image will be stored.
    :param width: The width of the montage (default=60000).
    :param level: Resolution downsampling level at which the WSI will be read (default=2).
    :param image_suffix: Suffix of image thumbnails (default=`.png`).
    :param masks: Flag to denote if masks need to be included (default=True).
    :param temp_dir: Optional path to temporary directory that stores the slide thumbnails.
        If `None`(default), a temporary directory will be created in `tmp` folder.
    :param cleanup: Flag to determine whether to clean the temporary directory containing thumbnails
        after the montage is created (default=False).
    :param num_parallel: Number of parallel processes for thumbnail creation. Use 0 or 1 to disable parallel.
    :param backend: The backend to use for reading the WSI (default=`cucim`).
    """
    # There might be some blanks at the bottom right
    # rows * cols <= N
    # We are going to stack the slides and their masks slide by side, so we need 2 * cols
    # 2 * cols / rows = 16 / 9; rows = 2 * cols * 9 / 16
    # cols * 2 * cols * 9 / 16 <= N; cols <= (N / 2 * 16 / 9)**(1 / 2)
    num_slides = len(records)
    multiplier = 2 if masks else 1
    num_cols = int(np.sqrt(num_slides / multiplier * 16 / 9))
    logging.info(f"Creating montage from {num_slides} slides with {num_cols} columns.")
    slide_width = (width // num_cols) // multiplier
    slide_size = slide_width, slide_width // 2
    temp_dir = tempfile.mkdtemp() if temp_dir is None else temp_dir
    temp_dir = Path(temp_dir)
    image_thumbnail_dir = temp_dir / "images"
    mask_thumbnail_dir = temp_dir / "masks" if masks else None

    if image_thumbnail_dir.is_dir():
        logging.info(f"Skipping thumbnail creation because folder already exists: {image_thumbnail_dir}")
    else:
        logging.info(f"Starting thumbnail creation with thumbnail size {slide_size}")
        make_thumbnails(
            records=records,
            slide_size=slide_size,
            images_dir=image_thumbnail_dir,
            level=level,
            masks_dir=mask_thumbnail_dir,
            image_suffix=image_suffix,
            num_parallel=num_parallel,
            backend=backend,
        )
    try:
        logging.info("Starting montage creation")
        montage_pil = make_montage_from_dir(
            image_thumbnail_dir, num_cols, masks_dir=mask_thumbnail_dir, image_suffix=image_suffix
        )
    except Exception as ex:
        raise ValueError(f"Failed to create montage from {image_thumbnail_dir}: {ex}")
    logger.info(f"Saving montage to {out_path}")
    montage_pil.save(out_path)
    if out_path.suffix != '.jpg':
        jpeg_out = out_path.with_suffix('.jpg')
        montage_pil.save(jpeg_out, format='JPEG', quality=90)
    if cleanup:
        shutil.rmtree(temp_dir)


class MontageCreation(MontageConfig):
    def read_list(self, csv_file_path: Optional[Path]) -> List[str]:
        """Reads a list of slide IDs from a file."""
        if csv_file_path:
            df = pd.read_csv(csv_file_path)
            column_to_read = df.columns[0]
            if len(df.columns) > 1:
                logger.warning(f"More than one column in file, using first column: {column_to_read}")
            return df[column_to_read].tolist()
        else:
            return []

    def read_exclusion_list(self) -> List[str]:
        """Read the list of slide IDs that should be excluded from the montage."""
        if self.exclude_by_slide_id:
            slides_to_exclude = self.read_list(self.exclude_by_slide_id)
            logger.info(f"Excluding {len(slides_to_exclude)} slides from montage. First 3: {slides_to_exclude[:3]}")
            logger.info(
                "Exclusion list will be matched against the Slide ID column (for predefined datasets) or the "
                "filename."
            )
            return slides_to_exclude
        else:
            return []

    def read_inclusion_list(self) -> List[str]:
        """Read the list of slide IDs that should be included in the montage."""
        if self.include_by_slide_id:
            slides_to_include = self.read_list(self.include_by_slide_id)
            logger.info(f"Restricting montage to {len(slides_to_include)} slides. First 3: {slides_to_include[:3]}")
            logger.info(
                "Inclusion list will be matched against the Slide ID column (for predefined datasets) or the "
                "filename."
            )
            return slides_to_include
        else:
            return []

    def read_dataset(self, input_folder: Path) -> DatasetOrDataframe:
        """Read the dataset that should be used for creating the montage. If a glob pattern has been provided, then
        all the image files specified by that pattern will be used for the montage. Otherwise, a file `dataset.csv`
        is expected in the input folder. The `dataset.csv` will be used to create an instance of `SlidesDataset`.

        :param input_folder: The folder where the dataset is located.
        :return: A SlidesDataset or dataframe object that contains the dataset."""
        if self.image_glob_pattern:
            logger.info(f"Trying to create a dataset from files that match: {self.image_glob_pattern}")
            try:
                dataset = dataset_from_folder(input_folder, glob_pattern=self.image_glob_pattern)
            except Exception as ex:
                raise ValueError(f"Unable to create dataset from files in folder {input_folder}: {ex}")
            if len(dataset) == 0:
                raise ValueError(f"No images found in folder {input_folder} with pattern {self.image_glob_pattern}")
            return dataset
        else:
            logger.info(f"Trying to load the dataset as a SlidesDataset from folder {input_folder}")
            try:
                dataset = SlidesDataset(root=input_folder)
            except Exception as ex:
                logging.error("Unable to load dataset.")
                file = input_folder / DEFAULT_DATASET_CSV
                # Print the whole directory tree to check where the problem is.
                while str(file) != str(file.root):
                    logging.debug(f"File: {file}, exists: {file.exists()}")
                    file = file.parent
                raise ValueError(
                    f"Unable to load dataset. Check if the file {DEFAULT_DATASET_CSV} "
                    f"exists, or provide a file name pattern via --image_glob_pattern. Error: {ex}"
                )
            return dataset

    def create_montage(self, input_folder: Path) -> None:
        """Creates a montage from the dataset in the input folder. The method reads the dataset, creates an output
        folder, handles the inclusion and exclusion lists, and then calls the method that creates the montage.

        :param input_folder: The folder where the dataset is located.
        :raises ValueError: If both an inclusion and exclusion list have been provided.
        """
        dataset = self.read_dataset(input_folder)
        self.output_path.mkdir(parents=True, exist_ok=True)
        if self.include_by_slide_id and self.exclude_by_slide_id:
            raise ValueError("You cannot provide both an inclusion and exclusion list.")
        if self.include_by_slide_id:
            items = self.read_inclusion_list()
            exclude_items = False
        elif self.exclude_by_slide_id:
            items = self.read_exclusion_list()
            exclude_items = True
        else:
            items = []
            exclude_items = True
        self.montage_from_included_and_excluded_slides(
            dataset=dataset,
            items=items,
            exclude_items=exclude_items,
        )

    def montage_from_included_and_excluded_slides(
        self,
        dataset: DatasetOrDataframe,
        items: Optional[List[str]] = None,
        exclude_items: bool = True,
        restrict_by_column: str = "",
    ) -> Optional[Path]:
        """Creates a montage of included and excluded slides from the dataset.

        :param dataset: Slides dataset or a plain dataframe.
        :param items: A list values for SlideID that should be included/excluded from the montage.
        :param exclude_items: If True, exclude the list in `items` from the montage. If False, include
            only those in the montage.
        :param restrict_by_column: The column name that should be used for inclusion/exclusion lists
            (default=dataset.slide_id_column).
        :return: A path to the created montage, or None if no images were available for creating the montage.
        """
        if isinstance(dataset, pd.DataFrame):
            df_original = dataset
        else:
            df_original = dataset.dataset_df
        logging.info(f"Input dataset contains {len(df_original)} records.")

        if restrict_by_column == "":
            if isinstance(dataset, pd.DataFrame):
                restrict_by_column = SlideKey.SLIDE_ID.value
            else:
                restrict_by_column = dataset.slide_id_column
        if items:
            if exclude_items:
                logging.info(f"Using dataset column '{restrict_by_column}' to exclude slides")
                include = False
            else:
                logging.info(f"Using dataset column '{restrict_by_column}' to restrict the set of slides")
                include = True
            df_restricted = restrict_dataset(df_original, column=restrict_by_column, items=items, include=include)
            logging.info(f"Updated dataset contains {len(df_restricted)} records")
        else:
            df_restricted = df_original

        montage_result = self.output_path / MONTAGE_FILE
        logging.info(f"Creating montage in {montage_result}")
        if isinstance(dataset, pd.DataFrame):
            records_restricted = dataset_to_records(df_restricted)
        else:
            dataset.dataset_df = df_restricted
            records_restricted = dataset_to_records(dataset)
            # We had to modify the dataset in place, hence restore the original dataset to avoid odd side-effects.
            dataset.dataset_df = df_original

        if len(records_restricted) > 0:
            make_montage(
                records=records_restricted,
                out_path=montage_result,
                width=self.width,
                level=self.level,
                masks=False,
                cleanup=True,
                num_parallel=self.parallel,
                backend=self.backend,
            )
            return montage_result
        else:
            logging.info("No slides to include in montage, skipping.")
            return None


def create_config_from_args() -> MontageConfig:
    """Creates a configuration object for montage creation from the commandline arguments.

    :return: An object that describes all options for the montage creation.
    """
    parser = create_montage_argparser()
    config = MontageCreation()
    parser_results = parse_arguments(parser, args=sys.argv[1:], fail_on_unknown_args=True)
    _ = apply_overrides(config, parser_results.args)
    return config



#  ------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License (MIT). See LICENSE in the repo root for license information.
#  ------------------------------------------------------------------------------------------
from argparse import ArgumentParser
from pathlib import Path
from typing import Optional

import param

from health_azure.argparsing import create_argparser


class AzureRunConfig(param.Parameterized):
    cluster: str = param.String(
        default="",
        allow_None=False,
        doc="The name of the GPU or CPU cluster inside the AzureML workspace"
        "that should execute the job. To run on your local machine, omit this argument.",
    )
    datastore: str = param.String(default="", doc="The name of the AzureML datastore where the dataset is defined.")
    dataset: str = param.String(
        default="",
        doc="The name of the AzureML dataset to use for creating the montage. The dataset will be "
        "mounted automatically. Use an absolute path to a folder on the local machine to bypass "
        "mounting.",
    )
    conda_env: Optional[Path] = param.ClassSelector(
        class_=Path,
        default=Path("hi-ml/hi-ml-cpath/environment.yml"),
        allow_None=True,
        doc="The Conda environment file that should be used when submitting the present run to "
        "AzureML. If not specified, the hi-ml-cpath environment file will be used.",
    )
    wait_for_completion: bool = param.Boolean(
        default=False,
        doc="If True, wait for AML Run to complete before proceeding. If False, submit the run to AML and exit",
    )
    docker_shm_size: str = param.String("100g", doc="The shared memory in the Docker image for the AzureML VMs.")
    workspace_config_path: Optional[Path] = param.ClassSelector(
        class_=Path,
        default=None,
        allow_None=True,
        doc="The path to the AzureML workspace configuration file. If not specified, the "
        "configuration file in the current folder or one of its parents will be used.",
    )
    display_name: str = param.String(
        default="", doc="The display name of the AzureML run. If not specified, a default name will be used."
    )


class MontageConfig(AzureRunConfig):
    level: int = param.Integer(
        default=1,
        doc="Resolution downsample level, e.g. if lowest resolution is 40x and the available "
        "downsample levels are [1.0, 4.0, 16.0] then level = 1 corresponds to 10x magnification",
    )
    exclude_by_slide_id: Optional[Path] = param.ClassSelector(
        class_=Path,
        default=None,
        allow_None=True,
        doc="Provide a file that contains slide IDs that should be excluded. File format is "
        "CSV, the first column is used as the slide ID. If the file is empty, no slides "
        "will be excluded.",
    )
    include_by_slide_id: Optional[Path] = param.ClassSelector(
        class_=Path,
        default=None,
        allow_None=True,
        doc="Provide a file that contains slide IDs that should be included. File format is "
        "CSV, the first column is used as the slide ID. If the file is empty, no montage "
        "will be produced.",
    )
    image_glob_pattern: str = param.String(
        default="",
        doc="When provided, use this pattern in rglob to find the files that should be included in the "
        "montage. Example: '**/*.tiff' to find all TIFF files recursive. You may have to escape "
        "the pattern in your shell.",
    )
    width: int = param.Integer(default=60_000, doc="The width of the montage in pixels")
    output_path: Path = param.ClassSelector(
        class_=Path, default=Path("outputs"), doc="The folder where the montage will be saved"
    )
    parallel: int = param.Integer(default=8, doc="The number of parallel processes to use when creating the montage.")
    backend: str = param.String(
        default="openslide", doc="The backend to use for reading the slides. Can be 'openslide' or 'cucim'"
    )


def create_montage_argparser() -> ArgumentParser:
    return create_argparser(
        MontageConfig(),
        usage="python create_montage.py --dataset <dataset_folder> --image_glob_pattern '**/*.tiff' --width 1000",
        description="Create an overview image with thumbnails of all slides in a dataset.",
    )



#  ------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License (MIT). See LICENSE in the repo root for license information.
#  ------------------------------------------------------------------------------------------

from pathlib import Path
from typing import Any, Dict, Optional, Tuple, Union

import numpy as np
import pandas as pd
import torch
from sklearn.utils.class_weight import compute_class_weight
from torch.utils.data import Dataset

from health_cpath.utils.naming import SlideKey, TileKey


DEFAULT_TRAIN_SPLIT_LABEL = "train"  # Value used to indicate the training split in `SPLIT_COLUMN`
DEFAULT_VAL_SPLIT_LABEL = "val"  # Value used to indicate the validation split in `SPLIT_COLUMN`
DEFAULT_TEST_SPLIT_LABEL = "test"  # Value used to indicate the test split in `SPLIT_COLUMN`
DEFAULT_LABEL_COLUMN = "label"  # Default column name for the label column


class TilesDataset(Dataset):
    """Base class for datasets of WSI tiles, iterating dictionaries of image paths and metadata.

    :param TILE_ID_COLUMN: CSV column name for tile ID.
    :param SLIDE_ID_COLUMN: CSV column name for slide ID.
    :param IMAGE_COLUMN: CSV column name for relative path to image file.
    :param PATH_COLUMN: CSV column name for relative path to image file. Replicated to propagate the path to the batch.
    :param SPLIT_COLUMN: CSV column name for train/test split (optional).
    :param TILE_X_COLUMN: CSV column name for horizontal tile coordinate (optional).
    :param TILE_Y_COLUMN: CSV column name for vertical tile coordinate (optional).
    :param DEFAULT_CSV_FILENAME: Default name of the dataset CSV at the dataset rood directory.
    """

    TILE_ID_COLUMN: str = 'tile_id'
    SLIDE_ID_COLUMN: str = 'slide_id'
    IMAGE_COLUMN: str = 'image'
    PATH_COLUMN: str = 'image_path'
    SPLIT_COLUMN: Optional[str] = 'split'
    TILE_X_COLUMN: Optional[str] = 'tile_x'
    TILE_Y_COLUMN: Optional[str] = 'tile_y'

    DEFAULT_CSV_FILENAME: str = "dataset.csv"

    def __init__(
        self,
        root: Union[str, Path],
        dataset_csv: Optional[Union[str, Path]] = None,
        dataset_df: Optional[pd.DataFrame] = None,
        train: Optional[bool] = None,
        validate_columns: bool = True,
        label_column: str = DEFAULT_LABEL_COLUMN,
        n_classes: int = 1,
        dataframe_kwargs: Dict[str, Any] = {},
    ) -> None:
        """
        :param root: Root directory of the dataset.
        :param dataset_csv: Full path to a dataset CSV file, containing at least
        `TILE_ID_COLUMN`, `SLIDE_ID_COLUMN`, and `IMAGE_COLUMN`. If omitted, the CSV will be read
        from `"{root}/{DEFAULT_CSV_FILENAME}"`.
        :param dataset_df: A potentially pre-processed dataframe in the same format as would be read
        from the dataset CSV file, e.g. after some filtering. If given, overrides `dataset_csv`.
        :param train: If `True`, loads only the training split (resp. `False` for test split). By
        default (`None`), loads the entire dataset as-is.
        :param validate_columns: Whether to call `validate_columns()` at the end of `__init__()`.
        `validate_columns()` checks that the loaded data frame for the dataset contains the expected column names
        for this class
        :param label_column: CSV column name for tile label. Defaults to `DEFAULT_LABEL_COLUMN="label"`.
        :param n_classes: Number of classes indexed in `label_column`. Default is 1 for binary classification.
        :param dataframe_kwargs: Keyword arguments to pass to `pd.read_csv()` when loading the dataset CSV.
        """
        if self.SPLIT_COLUMN is None and train is not None:
            raise ValueError("Train/test split was specified but dataset has no split column")

        self.root_dir = Path(root)
        self.label_column = label_column
        self.n_classes = n_classes

        if dataset_df is not None:
            self.dataset_csv = None
        else:
            self.dataset_csv = dataset_csv or self.root_dir / self.DEFAULT_CSV_FILENAME
            dataset_df = pd.read_csv(self.dataset_csv, **dataframe_kwargs)

        if dataset_df.index.name != self.TILE_ID_COLUMN:
            dataset_df = dataset_df.set_index(self.TILE_ID_COLUMN)
        if train is None:
            self.dataset_df = dataset_df
        else:
            split = DEFAULT_TRAIN_SPLIT_LABEL if train else DEFAULT_TEST_SPLIT_LABEL
            self.dataset_df = dataset_df[dataset_df[self.SPLIT_COLUMN] == split]

        if validate_columns:
            self.validate_columns()

    def validate_columns(self) -> None:
        """Check that loaded dataframe contains expected columns, raises `ValueError` otherwise.

        If the constructor is overloaded in a subclass, you can pass `validate_columns=False` and
        call `validate_columns()` after creating derived columns, for example.
        """
        columns = [
            self.SLIDE_ID_COLUMN,
            self.IMAGE_COLUMN,
            self.label_column,
            self.SPLIT_COLUMN,
            self.TILE_X_COLUMN,
            self.TILE_Y_COLUMN,
        ]
        columns_not_found = []
        for column in columns:
            if column is not None and column not in self.dataset_df.columns:
                columns_not_found.append(column)
        if len(columns_not_found) > 0:
            raise ValueError(f"Expected columns '{columns_not_found}' not found in the dataframe")

    def __len__(self) -> int:
        return self.dataset_df.shape[0]

    def __getitem__(self, index: int) -> Dict[str, Any]:
        tile_id = self.dataset_df.index[index]
        sample = {self.TILE_ID_COLUMN: tile_id, **self.dataset_df.loc[tile_id].to_dict()}
        sample[self.IMAGE_COLUMN] = str(self.root_dir / sample.pop(self.IMAGE_COLUMN))
        # we're replicating this column because we want to propagate the path to the batch
        sample[self.PATH_COLUMN] = sample[self.IMAGE_COLUMN]
        return sample

    @property
    def slide_ids(self) -> pd.Series:
        return self.dataset_df[self.SLIDE_ID_COLUMN]

    def get_slide_labels(self) -> pd.Series:
        return self.dataset_df.groupby(self.SLIDE_ID_COLUMN)[self.label_column].agg(pd.Series.mode)

    def get_class_weights(self) -> torch.Tensor:
        slide_labels = self.get_slide_labels()
        classes = np.unique(slide_labels)
        class_weights = compute_class_weight(class_weight='balanced', classes=classes, y=slide_labels)
        return torch.as_tensor(class_weights)

    def copy_coordinates_columns(self) -> None:
        """Copy columns "left" --> "tile_x" and "top" --> "tile_y" to be consistent with TilesDataset `TILE_X_COLUMN`
        and `TILE_Y_COLUMN`."""

        if TileKey.TILE_LEFT in self.dataset_df.columns:
            self.dataset_df = self.dataset_df.assign(**{TilesDataset.TILE_X_COLUMN: self.dataset_df[TileKey.TILE_LEFT]})
        if TileKey.TILE_TOP in self.dataset_df.columns:
            self.dataset_df = self.dataset_df.assign(**{TilesDataset.TILE_Y_COLUMN: self.dataset_df[TileKey.TILE_TOP]})


DEFAULT_DATASET_CSV = "dataset.csv"


class SlidesDataset(Dataset):
    """Base class for datasets of WSIs, iterating dictionaries of image paths and metadata.

    The output dictionaries are indexed by `..utils.naming.SlideKey`.
    """

    def __init__(
        self,
        root: Union[str, Path],
        dataset_csv: Optional[Union[str, Path]] = None,
        dataset_df: Optional[pd.DataFrame] = None,
        train: Optional[bool] = None,
        validate_columns: bool = True,
        label_column: str = DEFAULT_LABEL_COLUMN,
        n_classes: int = 1,
        dataframe_kwargs: Dict[str, Any] = {},
        default_csv_filename: str = DEFAULT_DATASET_CSV,
        slide_id_column: str = "slide_id",
        image_column: str = "image",
        mask_column: Optional[str] = None,
        split_column: Optional[str] = None,
        metadata_columns: Tuple[str, ...] = (),
    ) -> None:
        """
        :param root: Root directory of the dataset.
        :param dataset_csv: Full path to a dataset CSV file, containing at least
            `slide_id_column` and `image_column`. If omitted, the CSV will be read
            from `"{root}/{DEFAULT_CSV_FILENAME}"`.
        :param dataset_df: A potentially pre-processed dataframe in the same format as would be read
            from the dataset CSV file, e.g. after some filtering. If given, overrides `dataset_csv`.
        :param train: If `True`, loads only the training split (resp. `False` for test split). By
            default (`None`), loads the entire dataset as-is.
        :param validate_columns: Whether to call `validate_columns()` at the end of `__init__()`.
            `validate_columns()` checks that the loaded data frame for the dataset contains the expected column names
            for this class
        :param label_column: CSV column name for tile label. Default is `DEFAULT_LABEL_COLUMN="label"`.
        :param n_classes: Number of classes indexed in `label_column`. Default is 1 for binary classification.
        :param dataframe_kwargs: Keyword arguments to pass to `pd.read_csv()` when loading the dataset CSV.
        :param slide_id_column: CSV column name for slide ID. Default is `slide_id`.
        :param image_column: CSV column name for relative path to image file. Default is `image`.
        :param mask_column: CSV column name for relative path to mask file. Default is `None`.
        :param split_column: CSV column name for train/test split. Default is `None`.
        :param default_csv_filename: Default name of the dataset CSV at the dataset root directory.
        """
        self.root_dir = Path(root)
        self.label_column = label_column
        self.n_classes = n_classes
        self.dataframe_kwargs = dataframe_kwargs
        self.slide_id_column = slide_id_column
        self.image_column = image_column
        self.mask_column = mask_column
        self.split_column = split_column
        self.metadata_columns = metadata_columns
        self.default_csv_filename = default_csv_filename
        if self.split_column is None and train is not None:
            raise ValueError("Train/test split was specified but dataset has no split column")

        if dataset_df is not None:
            self.dataset_csv = None
        else:
            self.dataset_csv = dataset_csv or self.root_dir / self.default_csv_filename
            dataset_df = pd.read_csv(self.dataset_csv, **self.dataframe_kwargs)

        if dataset_df.index.name != self.slide_id_column:
            dataset_df = dataset_df.set_index(self.slide_id_column)
        if train is None:
            self.dataset_df = dataset_df
        else:
            split = DEFAULT_TRAIN_SPLIT_LABEL if train else DEFAULT_TEST_SPLIT_LABEL
            self.dataset_df = dataset_df[dataset_df[self.split_column] == split]

        if validate_columns:
            self.validate_columns()

    def validate_columns(self) -> None:
        """Check that loaded dataframe contains expected columns, raises `ValueError` otherwise.

        If the constructor is overloaded in a subclass, you can pass `validate_columns=False` and
        call `validate_columns()` after creating derived columns, for example.
        """
        mandatory_columns = {self.image_column, self.label_column, self.mask_column, self.split_column}
        optional_columns = (
            set(self.dataframe_kwargs["usecols"]) if "usecols" in self.dataframe_kwargs else set(self.metadata_columns)
        )
        columns = mandatory_columns.union(optional_columns)
        # slide_id_column is used for indexing and is not in df.columns anymore
        # None might be in columns if split_column is None
        columns_not_found = columns - set(self.dataset_df.columns) - {None, self.slide_id_column}
        if len(columns_not_found) > 0:
            raise ValueError(f"Expected columns '{columns_not_found}' not found in the dataframe")

    def __len__(self) -> int:
        return self.dataset_df.shape[0]

    def __getitem__(self, index: int) -> Dict[SlideKey, Any]:
        slide_id = self.dataset_df.index[index]
        slide_row = self.dataset_df.loc[slide_id]
        sample = {SlideKey.SLIDE_ID: slide_id}

        rel_image_path = slide_row[self.image_column]
        sample[SlideKey.IMAGE] = str(self.root_dir / rel_image_path)
        # we're replicating this column because we want to propagate the path to the batch
        sample[SlideKey.IMAGE_PATH] = sample[SlideKey.IMAGE]

        if self.mask_column:
            rel_mask_path = slide_row[self.mask_column]
            sample[SlideKey.MASK] = str(self.root_dir / rel_mask_path)
            sample[SlideKey.MASK_PATH] = sample[SlideKey.MASK]

        sample[SlideKey.LABEL] = slide_row[self.label_column]
        sample[SlideKey.METADATA] = {col: slide_row[col] for col in self.metadata_columns}
        return sample

    @classmethod
    def has_mask(cls) -> bool:
        return cls.MASK_COLUMN is not None

    def get_slide_labels(self) -> pd.Series:
        return self.dataset_df[self.label_column]

    def get_class_weights(self) -> torch.Tensor:
        slide_labels = self.get_slide_labels()
        classes = np.unique(slide_labels)
        class_weights = compute_class_weight(class_weight='balanced', classes=classes, y=slide_labels)
        return torch.as_tensor(class_weights)



#  ------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License (MIT). See LICENSE in the repo root for license information.
#  ------------------------------------------------------------------------------------------

import param
from torch import nn
from pathlib import Path
from typing import Optional, Tuple
from health_ml.utils.checkpoint_utils import CheckpointParser
from health_cpath.models.encoders import (
    DenseNet121_NoPreproc,
    HistoSSLEncoder,
    ImageNetSimCLREncoder,
    SSLEncoder,
    SwinTransformer_NoPreproc,
    TileEncoder,
    Resnet18,
    Resnet50,
    Resnet18_NoPreproc,
    Resnet50_NoPreproc,
)
from health_ml.networks.layers.attention_layers import (
    AttentionLayer,
    GatedAttentionLayer,
    MaxPoolingLayer,
    MeanPoolingLayer,
    TransformerPooling,
    TransformerPoolingBenchmark,
)


def set_module_gradients_enabled(model: nn.Module, tuning_flag: bool) -> None:
    """Given a model, enable or disable gradients for all parameters.

    :param model: A PyTorch model.
    :param tuning_flag: A boolean indicating whether to enable or disable gradients for the model parameters.
    """
    for params in model.parameters():
        params.requires_grad = tuning_flag


class EncoderParams(param.Parameterized):
    """Parameters class to group all encoder specific attributes for deepmil module."""

    encoder_type: str = param.String(doc="Name of the encoder class to use.")
    tile_size: int = param.Integer(default=224, bounds=(1, None), doc="Tile width/height, in pixels.")
    n_channels: int = param.Integer(default=3, bounds=(1, None), doc="Number of channels in the tile.")
    tune_encoder: bool = param.Boolean(
        False, doc="If True, fine-tune the encoder during training. If False (default), keep the encoder frozen."
    )
    pretrained_encoder = param.Boolean(
        False,
        doc="If True, transfer weights from the pretrained model (specified in `src_checkpoint`) to the encoder."
        "Else (False), keep the encoder weights as defined by the `encoder_type`.",
    )
    is_caching: bool = param.Boolean(
        default=False,
        doc="If True, cache the encoded tile features (disables random subsampling of tiles). If False (default), load "
        "the tiles without caching (enables random subsampling of tiles).",
    )
    encoding_chunk_size: int = param.Integer(
        default=0, doc="If > 0 performs encoding in chunks, by enconding_chunk_size tiles per chunk"
    )
    ssl_checkpoint: CheckpointParser = param.ClassSelector(
        class_=CheckpointParser, default=None, instantiate=False, doc=CheckpointParser.DOC
    )
    projection_dim: int = param.Integer(
        default=0, doc="If > 0, project the encoded tiles to this dimension. Otherwise, use identity projection."
    )
    use_encoder_checkpointing: bool = param.Boolean(
        default=False,
        doc="Whether to checkpoint activations during forward pass. This can be used to reduce the "
        "memory required to store gradients by checkpointing the activations. Note that this is only supported for "
        "Resnet18, Resnet50 and Swin Transformer encoders at the moment. The encoder class needs to override "
        "`custom_forward()` to support checkpointing. Default: False.",
    )
    checkpoint_segments_size: int = param.Integer(
        default=2,
        bounds=(1, None),
        doc="The segments size to use for checkpointing the encoder's activations. Default: 2. ",
    )
    batchnorm_momentum: Optional[float] = param.Number(
        default=None,
        bounds=(0, 1),
        doc="An optional momentum value to use for batch norm layers statistics updates when`use_encoder_checkpointing`"
        " is True. If None (default), sqrt of the default momentum retrieved from the model is used to avoid running "
        "statistics from going out of sync due to activations checkpointing.",
    )

    def validate(self) -> None:
        """Validate the encoder parameters."""
        if self.encoder_type == SSLEncoder.__name__ and not self.ssl_checkpoint:
            raise ValueError(
                "SSLEncoder requires an ssl_checkpoint. Please specify a valid checkpoint. "
                f"{CheckpointParser.INFO_MESSAGE}"
            )

    def get_encoder(self, outputs_folder: Optional[Path]) -> TileEncoder:
        """Given the current encoder parameters, returns the encoder object.

        :param outputs_folder: The output folder where SSL checkpoint should be saved.
        :raises ValueError: If the encoder type is not supported.
        :return: A TileEncoder instance for deepmil module.
        """
        encoder: TileEncoder
        if self.encoder_type == Resnet18.__name__:
            encoder = Resnet18(
                tile_size=self.tile_size,
                n_channels=self.n_channels,
                use_activation_checkpointing=self.use_encoder_checkpointing,
                checkpoint_segments_size=self.checkpoint_segments_size,
                batchnorm_momentum=self.batchnorm_momentum,
            )

        elif self.encoder_type == Resnet18_NoPreproc.__name__:
            encoder = Resnet18_NoPreproc(
                tile_size=self.tile_size,
                n_channels=self.n_channels,
                use_activation_checkpointing=self.use_encoder_checkpointing,
                checkpoint_segments_size=self.checkpoint_segments_size,
                batchnorm_momentum=self.batchnorm_momentum,
            )

        elif self.encoder_type == Resnet50.__name__:
            encoder = Resnet50(
                tile_size=self.tile_size,
                n_channels=self.n_channels,
                use_activation_checkpointing=self.use_encoder_checkpointing,
                checkpoint_segments_size=self.checkpoint_segments_size,
                batchnorm_momentum=self.batchnorm_momentum,
            )

        elif self.encoder_type == Resnet50_NoPreproc.__name__:
            encoder = Resnet50_NoPreproc(
                tile_size=self.tile_size,
                n_channels=self.n_channels,
                use_activation_checkpointing=self.use_encoder_checkpointing,
                checkpoint_segments_size=self.checkpoint_segments_size,
                batchnorm_momentum=self.batchnorm_momentum,
            )

        elif self.encoder_type == SwinTransformer_NoPreproc.__name__:
            encoder = SwinTransformer_NoPreproc(
                tile_size=self.tile_size,
                n_channels=self.n_channels,
                use_activation_checkpointing=self.use_encoder_checkpointing,
                checkpoint_segments_size=self.checkpoint_segments_size,
            )

        elif self.encoder_type == DenseNet121_NoPreproc.__name__:
            encoder = DenseNet121_NoPreproc(
                tile_size=self.tile_size,
                n_channels=self.n_channels,
                use_activation_checkpointing=self.use_encoder_checkpointing,
                checkpoint_segments_size=self.checkpoint_segments_size,
                batchnorm_momentum=self.batchnorm_momentum,
            )

        elif self.encoder_type == ImageNetSimCLREncoder.__name__:
            encoder = ImageNetSimCLREncoder(tile_size=self.tile_size, n_channels=self.n_channels)

        elif self.encoder_type == HistoSSLEncoder.__name__:
            encoder = HistoSSLEncoder(tile_size=self.tile_size, n_channels=self.n_channels)

        elif self.encoder_type == SSLEncoder.__name__:
            assert outputs_folder is not None, "outputs_folder cannot be None for SSLEncoder"
            encoder = SSLEncoder(
                pl_checkpoint_path=self.ssl_checkpoint.get_or_download_checkpoint(outputs_folder),
                tile_size=self.tile_size,
                n_channels=self.n_channels,
            )
        else:
            raise ValueError(f"Unsupported encoder type: {self.encoder_type}")
        set_module_gradients_enabled(encoder, tuning_flag=self.tune_encoder)
        return encoder

    def get_projection_layer(self, num_encoding: int) -> nn.Module:
        """If projection_dim > 0, returns a linear layer to project the encoded tiles to the projection_dim followed by
        relu activation. Else, returns an identity layer.

        :param num_encoding: The number of encoding dimensions.
        :return: A projection layer if projection_dim > 0, else Identity.
        """
        if self.projection_dim > 0:
            return nn.Sequential(nn.Linear(num_encoding, self.projection_dim), nn.ReLU())
        return nn.Identity()


class PoolingParams(param.Parameterized):
    """Parameters class to group all pooling specific attributes for deepmil module."""

    pool_type: str = param.String(doc="Name of the pooling layer class to use.")
    pool_hidden_dim: int = param.Integer(
        default=128,
        doc="If pooling has a learnable part, this defines the number of the hidden dimensions.",
    )
    pool_out_dim: int = param.Integer(1, doc="Dimension of the pooled representation.")
    num_transformer_pool_layers: int = param.Integer(
        default=4,
        doc="If transformer pooling is chosen, this defines the number of encoding layers.",
    )
    num_transformer_pool_heads: int = param.Integer(
        default=4,
        doc="If transformer pooling is chosen, this defines the number of attention heads.",
    )
    tune_pooling: bool = param.Boolean(
        default=True,
        doc="If True (default), fine-tune the pooling layer during training. If False, keep the pooling layer frozen.",
    )
    pretrained_pooling = param.Boolean(
        default=False,
        doc="If True, transfer weights from the pretrained model (specified in `src_checkpoint`) to the pooling"
        "layer. Else (False), initialize the pooling layer randomly.",
    )
    transformer_dropout: float = param.Number(
        default=0.0,
        doc="If transformer pooling is chosen, this defines the dropout of the tranformer encoder layers.",
    )

    def get_pooling_layer(self, num_encoding: int) -> Tuple[nn.Module, int]:
        """Given the pooling parameters, returns the pooling layer object.

        :param pooling_params: Pooling parameters that define the pooling layer class depending on the pooling type.
        :param num_encoding: The embedding dimension of the encoder.
        :raises ValueError: If the pooling type is not supported.
        :return: A tuple of (pooling layer, pooling output dimension) for deepmil module.
        """
        pooling_layer: nn.Module
        if self.pool_type == AttentionLayer.__name__:
            pooling_layer = AttentionLayer(
                input_dims=num_encoding, hidden_dims=self.pool_hidden_dim, attention_dims=self.pool_out_dim
            )
        elif self.pool_type == GatedAttentionLayer.__name__:
            pooling_layer = GatedAttentionLayer(
                input_dims=num_encoding, hidden_dims=self.pool_hidden_dim, attention_dims=self.pool_out_dim
            )
        elif self.pool_type == MeanPoolingLayer.__name__:
            pooling_layer = MeanPoolingLayer()
        elif self.pool_type == MaxPoolingLayer.__name__:
            pooling_layer = MaxPoolingLayer()
        elif self.pool_type == TransformerPooling.__name__:
            pooling_layer = TransformerPooling(
                num_layers=self.num_transformer_pool_layers,
                num_heads=self.num_transformer_pool_heads,
                dim_representation=num_encoding,
                transformer_dropout=self.transformer_dropout,
            )
            self.pool_out_dim = 1  # currently this is hardcoded in forward of the TransformerPooling
        elif self.pool_type == TransformerPoolingBenchmark.__name__:
            pooling_layer = TransformerPoolingBenchmark(
                num_layers=self.num_transformer_pool_layers,
                num_heads=self.num_transformer_pool_heads,
                dim_representation=num_encoding,
                hidden_dim=self.pool_hidden_dim,
                transformer_dropout=self.transformer_dropout,
            )
            self.pool_out_dim = 1  # currently this is hardcoded in forward of the TransformerPooling
        else:
            raise ValueError(f"Unsupported pooling type: {self.pool_type}")
        num_features = num_encoding * self.pool_out_dim
        set_module_gradients_enabled(pooling_layer, tuning_flag=self.tune_pooling)
        return pooling_layer, num_features


class ClassifierParams(param.Parameterized):
    dropout_rate: Optional[float] = param.Number(None, bounds=(0, 1), doc="Pre-classifier dropout rate.")
    tune_classifier: bool = param.Boolean(
        default=True,
        doc="If True (default), fine-tune the classifier during training. If False, keep the classifier frozen.",
    )
    pretrained_classifier: bool = param.Boolean(
        default=False,
        doc="If True, will use classifier weights from pretrained model specified in src_checkpoint. If False, will "
        "initiliaze classifier with random weights.",
    )

    def get_classifier(self, in_features: int, out_features: int) -> nn.Module:
        classifier_layer = nn.Linear(in_features=in_features, out_features=out_features)
        set_module_gradients_enabled(classifier_layer, tuning_flag=self.tune_classifier)
        if self.dropout_rate is None:
            return classifier_layer
        return nn.Sequential(nn.Dropout(self.dropout_rate), classifier_layer)



#  ------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License (MIT). See LICENSE in the repo root for license information.
#  ------------------------------------------------------------------------------------------

import functools
import shutil
import traceback
import warnings
from pathlib import Path
from typing import Any, Dict, Iterable, Optional, Tuple, Union

import numpy as np
import PIL
from monai.data import Dataset
from tqdm import tqdm
from health_ml.utils.box_utils import Box

from health_cpath.datasets.base_dataset import SlidesDataset
from health_cpath.preprocessing import tiling
from health_cpath.preprocessing.loading import LoadROId, WSIBackend, segment_foreground
from health_cpath.utils.naming import SlideKey, TileKey


def select_tiles(foreground_mask: np.ndarray, occupancy_threshold: float) -> Tuple[np.ndarray, np.ndarray]:
    """Exclude tiles that are mostly background based on estimated occupancy.

    :param foreground_mask: Boolean array of shape (*, H, W).
    :param occupancy_threshold: Tiles with lower occupancy (between 0 and 1) will be discarded.
    :return: A tuple containing which tiles were selected and the estimated occupancies. These will
    be boolean and float arrays of shape (*,), or scalars if `foreground_mask` is a single tile.
    """
    if occupancy_threshold < 0.0 or occupancy_threshold > 1.0:
        raise ValueError("Tile occupancy threshold must be between 0 and 1")
    occupancy = foreground_mask.mean(axis=(-2, -1))
    return (occupancy > occupancy_threshold).squeeze(), occupancy.squeeze()  # type: ignore


def get_tile_descriptor(tile_box: Box) -> str:
    """Format the box tile coordinates into a tile descriptor."""
    left, top = tile_box.x, tile_box.y
    right, bottom = left + tile_box.w, top + tile_box.h
    return f"left_{left:05d}_top_{top:05d}_right_{right:05d}_bottom_{bottom:05d}"


def get_tile_id(slide_id: str, tile_box: Box) -> str:
    """Format the slide ID and box tile coordinates into a unique tile ID."""
    return f"{slide_id}_{get_tile_descriptor(tile_box)}"


def save_image(array_chw: np.ndarray, path: Path) -> PIL.Image:
    """Save an image array in (C, H, W) format to disk."""
    path.parent.mkdir(parents=True, exist_ok=True)
    array_hwc = np.moveaxis(array_chw, 0, -1).astype(np.uint8).squeeze()
    pil_image = PIL.Image.fromarray(array_hwc)
    pil_image.convert('RGB').save(path)
    return pil_image


def generate_tiles(
    slide_image: np.ndarray, tile_size: int, foreground_threshold: float, occupancy_threshold: float
) -> Tuple[np.ndarray, np.ndarray, np.ndarray, int]:
    """Split the foreground of an input slide image into tiles.

    :param slide_image: The RGB image array in (C, H, W) format.
    :param tile_size: Lateral dimensions of each tile, in pixels.
    :param foreground_threshold: Luminance threshold (0 to 255) to determine tile occupancy.
    :param occupancy_threshold: Threshold (between 0 and 1) to determine empty tiles to discard.
    :return: A tuple containing the image tiles (N, C, H, W), tile coordinates (N, 2), occupancies
    (N,), and total number of discarded empty tiles.
    """
    image_tiles, tile_locations = tiling.tile_array_2d(slide_image, tile_size=tile_size, constant_values=255)
    foreground_mask, _ = segment_foreground(image_tiles, foreground_threshold)

    selected, occupancies = select_tiles(foreground_mask, occupancy_threshold)
    n_discarded = (~selected).sum()
    print(f"Percentage tiles discarded: {n_discarded / len(selected) * 100:.2f}")

    image_tiles = image_tiles[selected]
    tile_locations = tile_locations[selected]
    occupancies = occupancies[selected]

    return image_tiles, tile_locations, occupancies, n_discarded


def get_tile_info(sample: Dict[SlideKey, Any], occupancy: float, tile_box: Box, rel_slide_dir: Path) -> dict:
    """Map slide information and tiling outputs into tile-specific information dictionary.

    :param sample: Slide dictionary.
    :param occupancy: Estimated tile foreground occuppancy.
    :param tile_box: Tile box coordinates.
    :param rel_slide_dir: Directory where tiles are saved, relative to dataset root.
    :return: Tile information dictionary.
    """
    slide_id = sample[SlideKey.SLIDE_ID]
    descriptor = get_tile_descriptor(tile_box)
    rel_image_path = f"{rel_slide_dir}/{descriptor}.png"

    tile_info = {
        TileKey.SLIDE_ID: slide_id,
        TileKey.TILE_ID: get_tile_id(slide_id, tile_box),
        TileKey.IMAGE: rel_image_path,
        TileKey.LABEL: sample[SlideKey.LABEL],
        TileKey.TILE_LEFT.value: tile_box.x,
        TileKey.TILE_TOP.value: tile_box.y,
        TileKey.TILE_RIGHT.value: tile_box.x + tile_box.w,
        TileKey.TILE_BOTTOM.value: tile_box.y + tile_box.h,
        TileKey.OCCUPANCY: occupancy,
        TileKey.SLIDE_METADATA: {key: value for key, value in sample[SlideKey.METADATA].items()},
    }
    return tile_info


def format_csv_row(tile_info: Dict[TileKey, Any], keys_to_save: Iterable[TileKey], metadata_keys: Iterable[str]) -> str:
    """Format tile information dictionary as a row to write to a dataset CSV tile.

    :param tile_info: Tile information dictionary.
    :param keys_to_save: Which main keys to include in the row, and in which order.
    :param metadata_keys: Likewise for metadata keys.
    :return: The formatted CSV row.
    """
    tile_slide_metadata = tile_info.pop(TileKey.SLIDE_METADATA)

    fields = [str(tile_info[key]) for key in keys_to_save]
    fields.extend(str(tile_slide_metadata[key]) for key in metadata_keys)
    fields = ['"' + value + '"' if ',' in value else value for value in fields]  # if field contains a , add extra " "

    dataset_row = ','.join(fields)
    return dataset_row


def process_slide(
    sample: Dict[SlideKey, Any],
    level: int,
    margin: int,
    tile_size: int,
    foreground_threshold: Optional[float],
    occupancy_threshold: float,
    output_dir: Path,
    tile_progress: bool = False,
) -> None:
    """Load and process a slide, saving tile images and information to a CSV file.

    :param sample: Slide information dictionary, returned by the input slide dataset.
    :param level: Magnification level at which to process the slide.
    :param margin: Margin around the foreground bounding box, in pixels at lowest resolution.
    :param tile_size: Lateral dimensions of each tile, in pixels.
    :param foreground_threshold: Luminance threshold (0 to 255) to determine tile occupancy.
    If `None` (default), an optimal threshold will be estimated automatically.
    :param occupancy_threshold: Threshold (between 0 and 1) to determine empty tiles to discard.
    :param output_dir: Root directory for the output dataset; outputs for a single slide will be
    saved inside `output_dir/slide_id/`.
    :param tile_progress: Whether to display a progress bar in the terminal.
    """
    slide_metadata: Dict[str, Any] = sample[SlideKey.METADATA]
    keys_to_save = (
        TileKey.SLIDE_ID,
        TileKey.TILE_ID,
        TileKey.IMAGE,
        TileKey.LABEL,
        TileKey.TILE_LEFT,
        TileKey.TILE_TOP,
        TileKey.TILE_RIGHT,
        TileKey.TILE_BOTTOM,
        TileKey.OCCUPANCY,
    )
    metadata_keys = tuple(key for key in slide_metadata)
    csv_columns: Tuple[str, ...] = (*keys_to_save, *metadata_keys)

    slide_id: str = sample[SlideKey.SLIDE_ID]
    rel_slide_dir = Path(slide_id)
    slide_dir = output_dir / rel_slide_dir
    print(f">>> Slide dir {slide_dir}")
    if slide_dir.exists():  # already processed slide - skip
        print(f">>> Skipping {slide_dir} - already processed")
        return
    else:
        try:
            slide_dir.mkdir(parents=True)

            dataset_csv_path = slide_dir / "dataset.csv"
            dataset_csv_file = dataset_csv_path.open('w')
            dataset_csv_file.write(','.join(csv_columns) + '\n')  # write CSV header

            n_failed_tiles = 0
            failed_tiles_csv_path = slide_dir / "failed_tiles.csv"
            failed_tiles_file = failed_tiles_csv_path.open('w')
            failed_tiles_file.write('tile_id' + '\n')

            print(f"Loading slide {slide_id} ...")
            loader = LoadROId(
                backend=WSIBackend.CUCIM, level=level, margin=margin, foreground_threshold=foreground_threshold
            )
            sample = loader(sample)  # load 'image' from disk

            print(f"Tiling slide {slide_id} ...")
            image_tiles, rel_tile_locations, occupancies, _ = generate_tiles(
                sample[SlideKey.IMAGE], tile_size, sample[SlideKey.FOREGROUND_THRESHOLD], occupancy_threshold
            )

            tile_locations = (sample[SlideKey.SCALE] * rel_tile_locations + sample[SlideKey.ORIGIN]).astype(
                int
            )  # noqa: W503
            tile_size_scaled = int(tile_size * sample[SlideKey.SCALE])
            tile_boxes = [Box(x, y, tile_size_scaled, tile_size_scaled) for x, y in tile_locations]

            n_tiles = image_tiles.shape[0]

            print(f"Saving tiles for slide {slide_id} ...")
            for i in tqdm(range(n_tiles), f"Tiles ({slide_id[:6]}…)", unit="img", disable=not tile_progress):
                try:
                    tile_info = get_tile_info(sample, occupancies[i], tile_boxes[i], rel_slide_dir)
                    save_image(image_tiles[i], output_dir / tile_info[TileKey.IMAGE])
                    dataset_row = format_csv_row(tile_info, keys_to_save, metadata_keys)
                    dataset_csv_file.write(dataset_row + '\n')
                except Exception as e:
                    n_failed_tiles += 1
                    descriptor = get_tile_descriptor(tile_boxes[i])
                    failed_tiles_file.write(descriptor + '\n')
                    traceback.print_exc()
                    warnings.warn(
                        f"An error occurred while saving tile " f"{get_tile_id(slide_id, tile_boxes[i])}: {e}"
                    )

            dataset_csv_file.close()
            failed_tiles_file.close()
            if n_failed_tiles > 0:
                # TODO what we want to do with slides that have some failed tiles?
                print(f"{slide_id} is incomplete. {n_failed_tiles} tiles failed.")
            print(f"Finished processing slide {slide_id}")
        except Exception as e:
            traceback.print_exc()
            warnings.warn(f"An error occurred while processing slide {slide_id}: {e}")


def merge_dataset_csv_files(dataset_dir: Path) -> Path:
    """Combines all "*/dataset.csv" files into a single "dataset.csv" file in the given directory."""
    full_csv = dataset_dir / "dataset.csv"
    # TODO change how we retrieve these filenames, probably because mounted, the operation is slow
    #  and it seems to find many more files
    # print("List of files")
    # print([str(file) + '\n' for file in dataset_dir.glob("*/dataset.csv")])
    with full_csv.open('w') as full_csv_file:
        # full_csv_file.write(','.join(CSV_COLUMNS) + '\n')  # write CSV header
        first_file = True
        for slide_csv in tqdm(dataset_dir.glob("*/dataset.csv"), desc="Merging dataset.csv", unit='file'):
            print(f"Merging slide {slide_csv}")
            content = slide_csv.read_text()
            if not first_file:
                content = content[content.index('\n') + 1 :]  # discard header row for all but the first file
            full_csv_file.write(content)
            first_file = False
    return full_csv


def main(
    slides_dataset: SlidesDataset,
    root_output_dir: Union[str, Path],
    level: int,
    tile_size: int,
    margin: int,
    foreground_threshold: Optional[float],
    occupancy_threshold: float,
    parallel: bool = False,
    overwrite: bool = False,
    n_slides: Optional[int] = 0,
    only: Optional[str] = None,
) -> None:
    """Process a slides dataset to produce a tiles dataset.

    :param slides_dataset: Input tiles dataset object.
    :param root_output_dir: The root directory of the output tiles dataset.
    :param level: Magnification level at which to process the slide.
    :param tile_size: Lateral dimensions of each tile, in pixels.
    :param margin: Margin around the foreground bounding box, in pixels at lowest resolution.
    :param foreground_threshold: Luminance threshold (0 to 255) to determine tile occupancy.
    If `None` (default), an optimal threshold will be estimated automatically.
    :param occupancy_threshold: Threshold (between 0 and 1) to determine empty tiles to discard.
    :param parallel: Whether slides should be processed in parallel with multiprocessing.
    :param overwrite: Whether to overwrite an existing output tiles dataset. If `True`, will delete
    and recreate `root_output_dir`, otherwise will resume by skipping already processed slides.
    :param n_slides: If given, limit the total number of slides for debugging.
    :param only: Id of a single slide for debugging.
    """

    # Ignoring some types here because mypy is getting confused with the MONAI Dataset class
    # to select a subsample use keyword n_slides
    if only:
        slides_dataset.dataset_df = slides_dataset.dataset_df.filter(items=[only], axis=0)
    dataset = Dataset(slides_dataset)  # type: ignore

    if n_slides > 0:  # type: ignore
        dataset = dataset[:n_slides]

    output_dir = Path(root_output_dir)

    print(
        f"Creating dataset of level-{level} {tile_size}x{tile_size} "
        f"{slides_dataset.__class__.__name__} tiles at: {output_dir}"
    )

    if overwrite and output_dir.exists():
        shutil.rmtree(output_dir)
    output_dir.mkdir(parents=True, exist_ok=not overwrite)

    func = functools.partial(
        process_slide,
        level=level,
        margin=margin,
        tile_size=tile_size,
        foreground_threshold=foreground_threshold,
        occupancy_threshold=occupancy_threshold,
        output_dir=output_dir,
        tile_progress=not parallel,
    )

    if parallel:
        import multiprocessing

        pool = multiprocessing.Pool()
        map_func = pool.imap_unordered  # type: ignore
    else:
        map_func = map  # type: ignore

    list(tqdm(map_func(func, dataset), desc="Slides", unit="img", total=len(dataset)))  # type: ignore

    if parallel:
        pool.close()

    print("Merging slide files in a single file")
    merge_dataset_csv_files(output_dir)


if __name__ == '__main__':
    from health_cpath.datasets.tcga_prad_dataset import TcgaPradDataset

    # Example set up for an existing slides dataset:
    main(
        slides_dataset=TcgaPradDataset("/tmp/datasets/TCGA-PRAD_20220712"),
        root_output_dir="/tmp/datasets/TCGA-PRAD_10X_tiles_level1_224",
        n_slides=2,
        only=None,
        level=0,
        tile_size=224,
        margin=0,
        foreground_threshold=None,
        occupancy_threshold=0.05,
        parallel=True,
        overwrite=True,
    )



#  ------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License (MIT). See LICENSE in the repo root for license information.
#  ------------------------------------------------------------------------------------------
import logging
from pathlib import Path
from typing import Any, Collection, List, Optional, Sequence, Tuple, Dict

from sklearn.metrics import confusion_matrix
from torch import Tensor
import matplotlib.pyplot as plt
import numpy as np

from health_cpath.datasets.base_dataset import SlidesDataset
from health_cpath.preprocessing.loading import LoadingParams
from health_cpath.utils.viz_utils import (
    plot_attention_tiles,
    plot_heatmap_overlay,
    plot_attention_histogram,
    plot_normalized_and_non_normalized_confusion_matrices,
    plot_scores_hist,
    plot_slide,
)
from health_cpath.utils.analysis_plot_utils import plot_pr_curve, format_pr_or_roc_axes, plot_roc_curve
from health_cpath.utils.naming import PlotOption, ResultsKey, SlideKey
from health_cpath.utils.tiles_selection_utils import SlideNode, TilesSelector
from health_cpath.utils.viz_utils import save_figure


logger = logging.getLogger(__name__)

ResultsType = Dict[ResultsKey, List[Any]]
SlideDictType = Dict[SlideKey, Any]


def validate_class_names_for_plot_options(
    class_names: Optional[Sequence[str]], plot_options: Collection[PlotOption]
) -> Optional[Sequence[str]]:
    """Make sure that class names are provided if `PlotOption.CONFUSION_MATRIX` is among the chosen plot_options."""
    if PlotOption.CONFUSION_MATRIX in plot_options and not class_names:
        raise ValueError(
            "No class_names were provided while activating confusion matrix plotting. You need to specify the class "
            "names to use the `PlotOption.CONFUSION_MATRIX`"
        )
    return class_names


def save_scores_histogram(results: ResultsType, figures_dir: Path, stage: str = '') -> None:
    """Plots and saves histogram scores figure in its dedicated directory.

    :param results: Dict of lists that contains slide_level results
    :param figures_dir: The path to the directory where to save the histogram scores.
    :param stage: Test or validation, used to name the figure. Empty string by default.
    """
    fig = plot_scores_hist(results)
    save_figure(fig=fig, figpath=figures_dir / f"hist_scores_{stage}.png")


def save_pr_curve(
    results: ResultsType, figures_dir: Path, stage: str = '', stratify_metadata: Optional[List[Any]] = None
) -> None:
    """Plots and saves PR curve figure in its dedicated directory. This implementation
        only works for binary classification.
    ''
        :param results: Dict of lists that contains slide_level results
        :param figures_dir: The path to the directory where to save the figure
        :param stage: Test or validation, used to name the figure. Empty string by default.
    """
    true_labels = get_list_from_results_dict(results=results, results_key=ResultsKey.TRUE_LABEL)
    if len(set(true_labels)) == 2:
        scores = get_list_from_results_dict(results=results, results_key=ResultsKey.PROB)
        fig, ax = plt.subplots()
        plot_pr_curve(true_labels, scores, legend_label=stage, ax=ax)
        if stratify_metadata is not None:
            stratified_outputs = get_stratified_outputs(
                true_labels=true_labels, scores=scores, stratify_metadata=stratify_metadata
            )
            for key in stratified_outputs.keys():
                true_stratified = stratified_outputs[key][0]
                pred_stratified = stratified_outputs[key][1]
                plot_pr_curve(true_stratified, pred_stratified, legend_label=f"{stage}_{key}", ax=ax)
        ax.legend()
        format_pr_or_roc_axes(plot_type='pr', ax=ax)
        save_figure(fig=fig, figpath=figures_dir / f"pr_curve_{stage}.png")
    else:
        logger.warning("The PR curve plot implementation works only for binary cases, this plot will be skipped.")


def save_roc_curve(
    results: ResultsType, figures_dir: Path, stage: str = '', stratify_metadata: Optional[List[Any]] = None
) -> None:
    """Plots and saves ROC curve figure in its dedicated directory. This implementation
    only works for binary classification.

    :param results: Dict of lists that contains slide_level results
    :param figures_dir: The path to the directory where to save the figure
    :param stage: Test or validation, used to name the figure. Empty string by default.
    :param stratify_metadata: A list containing metadata values to plot stratified curves.
    """
    true_labels = get_list_from_results_dict(results=results, results_key=ResultsKey.TRUE_LABEL)
    if len(set(true_labels)) == 2:
        scores = get_list_from_results_dict(results=results, results_key=ResultsKey.PROB)
        fig, ax = plt.subplots()
        plot_roc_curve(true_labels, scores, legend_label=stage, ax=ax)
        if stratify_metadata is not None:
            stratified_outputs = get_stratified_outputs(
                true_labels=true_labels, scores=scores, stratify_metadata=stratify_metadata
            )
            for key in stratified_outputs.keys():
                true_stratified = stratified_outputs[key][0]
                pred_stratified = stratified_outputs[key][1]
                plot_roc_curve(true_stratified, pred_stratified, legend_label=f"{stage}_{key}", ax=ax)
        ax.legend()
        format_pr_or_roc_axes(plot_type='roc', ax=ax)
        save_figure(fig=fig, figpath=figures_dir / f"roc_curve_{stage}.png")
    else:
        logger.warning("The ROC curve plot implementation works only for binary cases, this plot will be skipped.")


def save_confusion_matrix(results: ResultsType, class_names: Sequence[str], figures_dir: Path, stage: str = '') -> None:
    """Plots and saves confusion matrix figure in its dedicated directory.

    :param results: Dict of lists that contains slide_level results
    :param class_names: List of class names.
    :param figures_dir: The path to the directory where to save the confusion matrix.
    :param stage: Test or validation, used to name the figure. Empty string by default.
    """
    true_labels = get_list_from_results_dict(results=results, results_key=ResultsKey.TRUE_LABEL)
    pred_labels = get_list_from_results_dict(results=results, results_key=ResultsKey.PRED_LABEL)
    all_potential_labels = list(range(len(class_names)))
    true_labels_diff_expected = set(true_labels).difference(set(all_potential_labels))
    pred_labels_diff_expected = set(pred_labels).difference(set(all_potential_labels))

    if true_labels_diff_expected != set():
        raise ValueError("More entries were found in true labels than are available in class names")
    if pred_labels_diff_expected != set():
        raise ValueError("More entries were found in predicted labels than are available in class names")

    cf_matrix = confusion_matrix(true_labels, pred_labels, labels=all_potential_labels)
    cf_matrix_n = confusion_matrix(true_labels, pred_labels, labels=all_potential_labels, normalize="true")

    fig = plot_normalized_and_non_normalized_confusion_matrices(cm=cf_matrix, cm_n=cf_matrix_n, class_names=class_names)
    save_figure(fig=fig, figpath=figures_dir / f"confusion_matrices_{stage}.png")


def save_top_and_bottom_tiles(
    case: str, slide_node: SlideNode, figures_dir: Path, num_columns: int = 4, figsize: Tuple[int, int] = (10, 10)
) -> None:
    """Plots and saves the top and bottom attention tiles of a given slide_node

    :param case: The report case (e.g., TP, FN, ...)
    :param slide_node: the slide_node for which we plot top and bottom tiles.
    :param figures_dir: The path to the directory where to save the attention tiles figure.
    """
    top_tiles_fig = plot_attention_tiles(
        slide_node=slide_node, case=case, top=True, num_columns=num_columns, figsize=figsize
    )
    save_figure(fig=top_tiles_fig, figpath=figures_dir / f"{slide_node.slide_id}_top.png")

    bottom_tiles_fig = plot_attention_tiles(
        slide_node=slide_node, case=case, top=False, num_columns=num_columns, figsize=figsize
    )
    save_figure(fig=bottom_tiles_fig, figpath=figures_dir / f"{slide_node.slide_id}_bottom.png")


def save_slide_thumbnail(case: str, slide_node: SlideNode, slide_dict: SlideDictType, figures_dir: Path) -> None:
    """Plots and saves a slide thumbnail

    :param case: The report case (e.g., TP, FN, ...)
    :param slide_node: The slide node that encapsulates the slide metadata.
    :param slide_dict: The slide dictionary that contains the slide image and other metadata.
    :param figures_dir: The path to the directory where to save the plots.
    """
    fig = plot_slide(case=case, slide_node=slide_node, slide_image=slide_dict[SlideKey.IMAGE], scale=1.0)
    save_figure(fig=fig, figpath=figures_dir / f"{slide_node.slide_id}_thumbnail.png")


def save_attention_heatmap(
    case: str,
    slide_node: SlideNode,
    slide_dict: SlideDictType,
    figures_dir: Path,
    results: ResultsType,
    tile_size: int = 224,
    should_upscale_coords: bool = False,
    extra_slide_dict: Optional[Dict[SlideKey, Any]] = None,
) -> None:
    """Plots and saves a slide thumbnail and attention heatmap

    :param case: The report case (e.g., TP, FN, ...)
    :param slide_node: The slide node that encapsulates the slide metadata.
    :param slide_dict: The slide dictionary that contains the slide image and other metadata.
    :param figures_dir: The path to the directory where to save the plots.
    :param results: Dict containing ResultsKey keys (e.g. slide id) and values as lists of output slides.
    :param tile_size: Size of each tile. Default 224.
    :param should_upscale_coords: Whether to upscale the coordinates of the attention heatmap. Default False.
    :param extra_slide_dict: An optional dictionary containing an extra slide image and metadata. Default None.
    """
    fig = plot_heatmap_overlay(
        case=case,
        slide_node=slide_node,
        slide_dict=slide_dict,
        results=results,
        tile_size=tile_size,
        should_upscale_coords=should_upscale_coords,
        extra_slide_dict=extra_slide_dict,
    )
    save_figure(fig=fig, figpath=figures_dir / f"{slide_node.slide_id}_heatmap.png")


def save_attention_histogram(case: str, slide_node: SlideNode, results: ResultsType, figures_dir: Path) -> None:
    """Plots a histogram of the attention values of the tiles in a bag.

    :param case: The report case (e.g., TP, FN, ...)
    :param slide_node: The slide node that encapsulates the slide metadata.
    :param results: Dict containing ResultsKey keys (e.g. slide id) and values as lists of output slides.
    :param figures_dir: The path to the directory where to save the plot.
    """
    fig = plot_attention_histogram(case=case, slide_node=slide_node, results=results)
    save_figure(fig=fig, figpath=figures_dir / f"{slide_node.slide_id}_histogram.png")


def make_figure_dirs(subfolder: str, parent_dir: Path) -> Path:
    """Create the figure directory"""
    figures_dir = parent_dir / subfolder
    figures_dir.mkdir(parents=True, exist_ok=True)
    return figures_dir


def get_list_from_results_dict(results: ResultsType, results_key: ResultsKey) -> List[Any]:
    """Get a specific results list from the slide_level results dictionary, we extract items from tensors
    here so that it's compatible with inputs formats of scikit learn functions.

    :param results: Dict of lists that contains slide_level results
    :param results_key: ResultsKey key for the list to be retrieved
    """
    return [i.item() if isinstance(i, Tensor) else i for i in results[results_key]]


def get_stratified_outputs(
    true_labels: List[Any], scores: List[Any], stratify_metadata: List[Any]
) -> Dict[str, List[Any]]:
    """
    Get stratified true labels and predictions given metadata, from all true and prediction labels
    to plot stratified curves.

    :param true_labels: list of true labels.
    :param scores: list of prediction scores.
    :param stratify_metadata: list containing the corresponding metadata values on which to stratify results.
    :return: A dictionary of stratified outputs, where a key is a unique value from the metadata,
    and value contains a list of two lists - true labels in first list, predicted labels in second list.
    """
    unique_vals, unique_counts = np.unique(stratify_metadata, return_counts=True)  # metadata should not contain nans
    stratified_outputs = {}
    for val, count in zip(unique_vals, unique_counts):
        idxs = [i for i, x in enumerate(stratify_metadata) if x == val]
        assert len(idxs) == count
        true_stratified = [true_labels[i] for i in idxs]
        pred_stratified = [scores[i] for i in idxs]
        stratified_outputs[val] = [true_stratified, pred_stratified]
    return stratified_outputs


class DeepMILPlotsHandler:
    def __init__(
        self,
        plot_options: Collection[PlotOption],
        loading_params: LoadingParams,
        tile_size: int = 224,
        num_columns: int = 4,
        figsize: Tuple[int, int] = (10, 10),
        stage: str = '',
        class_names: Optional[Sequence[str]] = None,
        stratify_plots_by: Optional[str] = None,
    ) -> None:
        """Class that handles the plotting of DeepMIL results.

        :param plot_options: A set of plot options to produce the desired plot outputs.
        :param loading_params: The loading parameters to use when loading the whole slide images.
        :param tile_size: The size of the tiles to use when plotting the attention tiles, defaults to 224
        :param num_columns: Number of columns to create the subfigures grid, defaults to 4
        :param figsize: The figure size of tiles attention plots, defaults to (10, 10)
        :param stage: Test or Validation, used to name the plots
        :param class_names: List of class names, defaults to None
        :param stratify_plots_by: Name of metadata field to stratify output plots (PR curve, ROC curve).
        `None` by default (no stratification).
        """

        self.plot_options = plot_options
        self.class_names = validate_class_names_for_plot_options(class_names, plot_options)
        self.tile_size = tile_size
        self.num_columns = num_columns
        self.figsize = figsize
        self.stage = stage
        self.loading_params = loading_params
        self.should_upscale_coords = loading_params.should_upscale_coordinates()
        self.loading_params.set_roi_type_to_foreground()
        self.slides_dataset: Optional[SlidesDataset] = None
        self.extra_slides_dataset: Optional[SlidesDataset] = None
        self.stratify_plots_by = stratify_plots_by

    def get_slide_dict(self, slide_node: SlideNode, slides_dataset: SlidesDataset) -> Optional[SlideDictType]:
        """Returns the slide dictionary for a given slide node from a slides dataset.

        :param slide_node: The slide node that encapsulates the slide metadata.
        :param slides_dataset: The slides dataset that contains the slide image and other metadata.
        """
        try:
            slide_index = slides_dataset.dataset_df.index.get_loc(slide_node.slide_id)
        except KeyError:
            logger.warning(f"Could not find slide {slide_node.slide_id} in the dataset. Skipping extra slide...")
            return None
        assert isinstance(slide_index, int), f"Got non-unique slide ID: {slide_node.slide_id}"
        slide_dict = slides_dataset[slide_index]
        loader = self.loading_params.get_load_roid_transform()
        slide_dict = loader(slide_dict)
        return slide_dict

    def save_slide_node_figures(
        self, case: str, slide_node: SlideNode, outputs_dir: Path, results: ResultsType
    ) -> None:
        """Plots and saves all slide related figures: `TOP_BOTTOM_TILES`, `SLIDE_THUMBNAIL` and `ATTENTION_HEATMAP`."""
        case_dir = make_figure_dirs(subfolder=case, parent_dir=outputs_dir)

        if PlotOption.TOP_BOTTOM_TILES in self.plot_options:
            save_top_and_bottom_tiles(case, slide_node, case_dir, self.num_columns, self.figsize)

        if PlotOption.ATTENTION_HISTOGRAM in self.plot_options:
            save_attention_histogram(case, slide_node, results, case_dir)

        if PlotOption.ATTENTION_HEATMAP in self.plot_options or PlotOption.SLIDE_THUMBNAIL in self.plot_options:
            assert self.slides_dataset is not None, "Cannot plot attention heatmap or thumbnail without slides dataset"
            slide_dict = self.get_slide_dict(slide_node=slide_node, slides_dataset=self.slides_dataset)
            assert slide_dict is not None, "Slide dict is None. Cannot plot attention heatmap or thumbnail."

            if PlotOption.SLIDE_THUMBNAIL in self.plot_options:
                save_slide_thumbnail(case=case, slide_node=slide_node, slide_dict=slide_dict, figures_dir=case_dir)

            if PlotOption.ATTENTION_HEATMAP in self.plot_options:
                if self.extra_slides_dataset is not None:
                    extra_slide_dict = self.get_slide_dict(slide_node, self.extra_slides_dataset)
                else:
                    extra_slide_dict = None
                save_attention_heatmap(
                    case,
                    slide_node,
                    slide_dict,
                    case_dir,
                    results,
                    self.tile_size,
                    self.should_upscale_coords,
                    extra_slide_dict,
                )

    def get_metadata(self, results: ResultsType) -> Optional[List[Any]]:
        """
        Get metadata of outputs (validation or test) from slides dataset to stratify plots (e.g PR curve, ROC curve).
        Returns metadata values of the results slides specified in `stratify_plots by` from slides dataset.
        Returns `None` if slides dataset is `None` or if `stratify_plots_by` is set to `None` (no stratification).

        :param results: Dict containing ResultsKey keys (e.g. slide id) and values as lists of output slides.
        """
        if self.slides_dataset is None or self.stratify_plots_by is None:
            stratify_metadata = None
            if self.stratify_plots_by is not None:
                logger.warning("Slides dataset is missing so stratified plots will be skipped.")
        else:
            slides_df = self.slides_dataset.dataset_df
            all_slide_ids = slides_df.index.to_list()
            output_slide_ids = [x[0] for x in results[ResultsKey.SLIDE_ID]]  # get unique slide ID from bags
            stratify_metadata = []
            for slide in output_slide_ids:
                idx = all_slide_ids.index(slide)
                sample = self.slides_dataset[idx]
                if self.stratify_plots_by not in sample[SlideKey.METADATA]:
                    logger.warning(
                        f"{self.stratify_plots_by} not available in the slides dataset metadata, \
                        make sure the dataset includes stratify_plots_by, so stratified plots will be skipped."
                    )
                    return None
                stratify_metadata.append(sample[SlideKey.METADATA][self.stratify_plots_by])
        return stratify_metadata

    def save_plots(self, outputs_dir: Path, tiles_selector: Optional[TilesSelector], results: ResultsType) -> None:
        """Plots and saves all selected plot options during inference (validation or test) time.

        :param outputs_dir: The root output directory where to save plots figures.
        :param tiles_selector: The tiles selector used to select top and bottom tiles from top and bottom slides.
        :param results: A dictionary of the validation or tests results.
        :param stage: The model stage: validation or test.
        """
        if self.plot_options:
            logger.info(f"Plotting {[opt.value for opt in self.plot_options]}...")
            figures_dir = make_figure_dirs(subfolder="fig", parent_dir=outputs_dir)

            stratify_metadata = self.get_metadata(results=results)

            if PlotOption.PR_CURVE in self.plot_options:
                save_pr_curve(
                    results=results, figures_dir=figures_dir, stage=self.stage, stratify_metadata=stratify_metadata
                )

            if PlotOption.ROC_CURVE in self.plot_options:
                save_roc_curve(
                    results=results, figures_dir=figures_dir, stage=self.stage, stratify_metadata=stratify_metadata
                )

            if PlotOption.HISTOGRAM in self.plot_options:
                save_scores_histogram(
                    results=results,
                    figures_dir=figures_dir,
                    stage=self.stage,
                )

            if PlotOption.CONFUSION_MATRIX in self.plot_options:
                assert self.class_names
                save_confusion_matrix(results, class_names=self.class_names, figures_dir=figures_dir, stage=self.stage)

            if tiles_selector:
                for class_id in range(tiles_selector.n_classes):
                    for slide_node in tiles_selector.top_slides_heaps[class_id]:
                        case = "TN" if class_id == 0 else f"TP_{class_id}"
                        self.save_slide_node_figures(case, slide_node, outputs_dir, results)

                    for slide_node in tiles_selector.bottom_slides_heaps[class_id]:
                        case = "FP" if class_id == 0 else f"FN_{class_id}"
                        self.save_slide_node_figures(case, slide_node, outputs_dir, results)



#  ------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License (MIT). See LICENSE in the repo root for license information.
#  ------------------------------------------------------------------------------------------

"""
Tools to upload training results from Azure Machine Learning runs to a deployed instance of the
`Digital Slide Archive <https://digitalslidearchive.github.io/>`_.

The `Girder Python client <https://girder.readthedocs.io/en/latest/python-client.html>`_ is used
to communicate with the API.

An example API can be found at the `Kitware demo <https://demo.kitware.com/histomicstk/api/v1>`_.
"""

from __future__ import annotations

import os
import sys
import logging
import argparse
import tempfile
import webbrowser
from pathlib import Path
from urllib.parse import urljoin
from dataclasses import dataclass, astuple, asdict
from typing import Any, Dict, List, Optional, Sequence, Union

import azureml
import numpy as np
import pandas as pd
from tqdm import tqdm
import matplotlib.pyplot as plt
from girder_client import GirderClient, HttpError
from health_azure.logging import logging_to_stdout
from health_azure.utils import get_aml_run_from_run_id

from health_cpath.utils.naming import ResultsKey
from health_cpath.utils.output_utils import (
    AML_OUTPUTS_DIR,
    EXTRA_VAL_OUTPUTS_SUBDIR,
    OUTPUTS_CSV_FILENAME,
    TEST_OUTPUTS_SUBDIR,
    VAL_OUTPUTS_SUBDIR,
)


TypeRectangleJSON = Dict[str, Union[str, float, Dict[str, str]]]
TypePointJSON = Dict[str, Union[str, List[float], Dict[str, str]]]
TypeAnnotationJSON = Dict[str, Union[str, List[Dict]]]

# DSA coordinates are generally expressed using three dimensions
Z_ZERO = [0]


@dataclass
class Color:
    """Container for RGBA color.

    All values are expected to be in :math:`[0, 255]`. Check the
    `documentation <https://github.com/girder/large_image/blob/master/girder_annotation/docs/annotations.rst#colors>`_
    for more information.
    """

    red: int
    green: int
    blue: int
    alpha: int

    def __post_init__(self) -> None:
        array = np.asarray(self)
        if (array < 0).any() or (array > 255).any():
            raise ValueError(f"All RGBA components must be between 0 and 255, but {astuple(self)} were passed")

    def __array__(self) -> np.ndarray:
        return np.array(self.components, dtype=int)

    @property
    def components(self) -> Sequence[int]:
        """Tuple of R, G, B and A components."""
        return astuple(self)

    def __str__(self) -> str:
        return f"rgba{astuple(self)}"


TRANSPARENT = Color(0, 0, 0, 0)


@dataclass
class Element:
    """Base class for annotations elements such as points or rectangles."""

    label: str
    fill_color: Color
    line_color: Color

    def as_json(self) -> Dict:
        """Return JSON representation suitable for the DSA."""
        raise NotImplementedError


@dataclass
class Coordinates:
    """Helper class to represent x and y coordinates."""

    x: float
    y: float

    def __post_init__(self) -> None:
        try:
            float(self.x)
            float(self.y)
        except ValueError as e:
            raise TypeError(f"Error converting coordinates to float: \"{asdict(self)}\"") from e

    def __array__(self) -> np.ndarray:
        return np.asarray(astuple(self))


@dataclass
class Rectangle(Element):
    """Container for rectangles in annotations.

    More information can be found in the
    `DSA documentation <https://github.com/girder/large_image/blob/master/girder_annotation/docs/annotations.rst#rectangle>`_.
    """  # noqa: E501

    left: float
    top: float
    right: float
    bottom: float

    def __post_init__(self) -> None:
        if self.left >= self.right:
            raise ValueError(f"The value for right ({self.right}) must be larger than the value for left ({self.left})")
        if self.top >= self.bottom:
            raise ValueError(f"The value for bottom ({self.bottom}) must be larger than the value for top ({self.top})")

    @property
    def width(self) -> float:
        return self.right - self.left

    @property
    def height(self) -> float:
        return self.bottom - self.top

    @property
    def center_xy(self) -> np.ndarray:
        size_xy = np.asarray((self.width, self.height))
        left_top = self.left, self.top
        return np.asarray(left_top) + size_xy / 2

    def as_json(self) -> TypeRectangleJSON:
        data: TypeRectangleJSON = {}
        data["fillColor"] = str(self.fill_color)
        data["lineColor"] = str(self.line_color)
        data["type"] = "rectangle"
        data["center"] = self.center_xy.tolist() + Z_ZERO
        data["width"] = float(self.width)
        data["height"] = float(self.height)
        data["label"] = {"value": self.label}
        return data


@dataclass
class Point(Element):
    """Container for points in DSA annotations.

    More information can be found in the
    `DSA documentation <https://github.com/girder/large_image/blob/master/girder_annotation/docs/annotations.rst#point>`_.
    """  # noqa: E501

    center: Coordinates

    def __post_init__(self) -> None:
        if not isinstance(self.center, Coordinates):
            raise TypeError(f"Center must be an instance of Coordinates, not {type(self.center)}")

    def as_json(self) -> TypePointJSON:
        data: TypePointJSON = {}
        data["fillColor"] = str(self.fill_color)
        data["lineColor"] = str(self.line_color)
        data["type"] = "point"
        data["center"] = list(astuple(self.center)) + Z_ZERO
        data["label"] = {"value": self.label}
        return data


@dataclass
class Annotation:
    """Container for DSA annotation.

    An example can be found in the
    `DSA documentation <https://github.com/girder/large_image/blob/master/girder_annotation/docs/annotations.rst#a-sample-annotation>`_.
    """  # noqa: B950

    name: str
    elements: Sequence[Element]
    description: str = ""

    def __post_init__(self) -> None:
        if not self.name:
            # This is enforced by the JSON schema
            raise ValueError("The annotation name cannot be empty")

    def as_json(self) -> TypeAnnotationJSON:
        data: TypeAnnotationJSON = {}
        data["name"] = self.name
        data["description"] = self.description
        data["elements"] = [element.as_json() for element in self.elements]  # type: ignore
        return data


class DigitalSlideArchive:
    """Representation of a deployed instance of the Digital Slide Archive.

    :param url: URL of a deployed instance of the `Digital Slide Archive <https://digitalslidearchive.github.io/>`_.
        For example: https://demo.kitware.com/histomicstk/. If not given, it must be defined in an environment
        variable ``DSA_URL``.
    :param api_key: Girder `API key <https://girder.readthedocs.io/en/latest/user-guide.html#api-keys>`_ to
        perform allowed operations. If not given, it must be defined in an environment
        variable ``DSA_API_KEY``.
    """

    URL_ENV_NAME = "DSA_URL"
    API_KEY_ENV_NAME = "DSA_API_KEY"

    def __init__(
        self,
        url: Optional[str] = None,
        api_key: Optional[str] = None,
    ):
        try:
            url = os.environ[self.URL_ENV_NAME] if url is None else url
        except KeyError as e:
            message = (
                "The DSA URL must be passed as an argument"
                f" or stored in an environment variable \"{self.URL_ENV_NAME}\""
            )
            raise RuntimeError(message) from e
        self.url = url
        self._client = self._get_client(self.api_url, api_key)
        self.post = self._client.post
        self.get = self._client.get

    def _get_client(self, api_url: str, api_key: Optional[str]) -> GirderClient:
        logging.info("Logging into DSA API hosted at %s...", api_url)
        client = GirderClient(apiUrl=api_url)
        try:
            api_key = os.environ[self.API_KEY_ENV_NAME] if api_key is None else api_key
        except KeyError as e:
            message = (
                "The DSA API key must be passed as an argument"
                f" or stored in an environment variable \"{self.API_KEY_ENV_NAME}\""
            )
            raise RuntimeError(message) from e
        client.authenticate(apiKey=api_key)
        logging.info("Authentication successful")
        return client

    @property
    def api_url(self) -> str:
        return urljoin(self.url, GirderClient.DEFAULT_API_ROOT)

    def add_annotation(self, item_id: str, annotation: Annotation) -> Dict:
        """Add annotation to a DSA item.

        :param item_id: Unique string representing the item ID.
        :param annotation: Instance of :class:`Annotation` that will be added to the item.
        """
        response = self.post(
            path="annotation",
            parameters={"itemId": item_id},
            json=annotation.as_json(),
        )
        return response

    def make_api_call(self, api_path: str, parameters: Dict, result_field: str = "") -> Any:
        """Issues a GET call to the DSA API on the given API path. If that raises a HttpError, a ValueError is raised
        with more details.

        :param api_path: The API path (for example, "resource/lookup")
        :param parameters: A parameter dictionary with parameters for the API call.
        :param result_field: If given, return only the specific item from the result dictionary.
        :raises ValueError: If a HttpError is raised during the call.
        :raises KeyError: If a specific field from the result was requested, but it was not found.
        :return: The result of the API call, or only a given field thereof if `result_field` was provided.
        """
        try:
            result = self.get(api_path, parameters=parameters)
        except HttpError as ex:
            raise ValueError(
                f"API '{api_path}' does not exist or cannot be accessed. Please check "
                f"the path and the access permissions of your API key. Exception: {ex}"
            )
        if result_field:
            if result_field in result:
                return result[result_field]
            raise KeyError(f"Item '{result_field}' is not present in result: {result}")
        return result

    def search_item(self, text: str, search_mode: str = "text") -> Item:
        """Search a file in the DSA collections and return its parent item.

        :param text: Text query.
        :param search_mode: Girder search mode. It ``'text'``, the full item ID will be matched.
            If ``'prefix'``, the file whose name starts with the value in ``text`` will be returned.
        :raises RuntimeError: If no items are found for the query or if more than one item is found.
        """
        parameters = dict(q=text, types="[\"item\"]", mode=search_mode)
        items_jsons = self.make_api_call("resource/search", parameters=parameters, result_field="item")
        if not items_jsons:
            raise RuntimeError(f"No items found for query \"{text}\"")
        elif len(items_jsons) > 1:
            raise RuntimeError(f"More than one item found for query \"{text}\":\n{items_jsons}")
        return Item(self, json=items_jsons[0])

    def get_folder_id(self, folder_path: str) -> str:
        """Retrieves the ID of a folder, when given a folder path.

        :param folder_path: The path of the folder to retrieve (for example, "Coll1/folder2" for `folder2` in
            collection `Coll1`)
        :return: The ID of the folder in DSA.
        """
        parameters = dict(path=f"/collection/{folder_path.strip('/')}")
        return self.make_api_call("resource/lookup", parameters=parameters, result_field="_id")

    def get_items_in_folder(self, folder_id: str) -> List[Item]:
        """Retrieves all items in the given folder.

        :param folder_id: The ID of the folder to read from.
        :return: A list of item JSON objects.
        """
        parameters = dict(type="folder", limit=5000)
        result = self.make_api_call(f"resource/{folder_id}/items", parameters=parameters)
        return [Item(self, json=item_json) for item_json in result]


class Item:
    """Representation of an item in the Digital Slide Archive.

    :param dsa: Instance of :class:`DigitalSlideArchive`.
    :param id: ID of the item in the Digital Slide Archive.
    :param json: JSON representation of the Digital Slide Archive item.
    :raises ValueError: If no ID or JSON is passed.
    :raises ValueError: If both an ID and JSON are passed.
    """

    def __init__(self, dsa: DigitalSlideArchive, id: Optional[str] = None, json: Optional[Dict] = None):
        self._dsa = dsa

        # Validate that only one of ID or the JSON dict have been passed
        if id is not None and json is not None:
            raise ValueError("Only the ID or the JSON object can be passed")
        if id is None and json is None:
            raise ValueError("An ID or JSON object must be passed")

        if id is None:
            id = json["_id"]  # type: ignore
        self.id = id
        self._json = json

    @property
    def name(self) -> str:
        """Gets the name of the item as stored in the "name" field in the JSON representation.
        If the "name" field is not present, return an empty string.

        :return: The "name" field of the JSON representation, or "" if that field is not present.
        """
        if self._json is None:
            return ""
        return self._json.get("name", "")

    @property
    def url(self) -> str:
        return urljoin(self._dsa.url, f"/#item/{self.id}")

    def open(self) -> bool:
        return webbrowser.open(self.url)

    def add_annotation(self, annotation: Annotation) -> Dict:
        response = self._dsa.post(
            path="annotation",
            parameters={"itemId": self.id},
            json=annotation.as_json(),
        )
        return response


class RunOutputs:
    """Class to process outputs CSV of an Azure Machine Learning (AML) run.

    :param run_id: ID of the AML run from which results will be downloaded.
    :param workspace_config_path: Path to an AML workspace configuration file, e.g., ``config.json``.
    :param overwrite_csv: Force download of the output CSV even when it is found locally.
    """

    def __init__(
        self,
        run_id: str,
        workspace_config_path: Optional[Path] = None,
        overwrite_csv: bool = False,
        split: str = "test",
    ):
        logging.info("Getting run \"%s\"...", run_id)
        run = get_aml_run_from_run_id(run_id, workspace_config_path=workspace_config_path)
        experiment = run.experiment
        workspace = experiment.workspace

        self.run = run
        self.experiment = experiment
        self.workspace = workspace
        self.df = self.get_df(overwrite_csv)
        self.tile_size = None
        self.split = split

    def get_df(self, overwrite_csv: bool) -> pd.DataFrame:
        """Download outputs CSV from Azure ML and read the data frame.

        The CSV is cached locally for future

        :param overwrite_csv: Force download of the output CSV even when it is found locally.
        """
        csv_filename = "/".join([AML_OUTPUTS_DIR, self.split, OUTPUTS_CSV_FILENAME])
        csv_stem = Path(csv_filename).stem
        csv_name = f"{csv_stem}-{self.workspace.name}-{self.run.id}.csv"
        cached_csv_path = Path(tempfile.gettempdir()) / csv_name
        if cached_csv_path.is_file() and not overwrite_csv:
            logging.info(f"Found cached CSV file {cached_csv_path}")
        else:
            logging.info("Downloading outputs CSV...")
            aml_exceptions = (
                azureml.exceptions._azureml_exception.UserErrorException,
                azureml._restclient.models.error_response.ErrorResponseException,
            )
            try:
                logging.info(f"Downloading file {csv_filename}")
                self.run.download_file(csv_filename, cached_csv_path)
            except aml_exceptions as e:
                raise FileNotFoundError("Error downloading outputs file from run") from e
        logging.info("Reading CSV file: %s ...", cached_csv_path)
        return self._read_csv(cached_csv_path)

    def get_annotation_from_slide_data_frame(
        self,
        df: pd.DataFrame,
        name: str,
        rescale: bool = False,
        colormap_name: str = "Greens",
        description: str = "",
    ) -> Annotation:
        """Create an annotation from a slide data frame.

        :param df: Data frame with data from a single slide.
        :param name: Annotation name.
        :param rescale: If ``True``, attention values will be remapped to :math:`[0, 1]` to increase
            the dynamic range of output colors.
        :param colormap_name: Name of the Matplotlib colormap used for the annotation elements.
        :param description: Optional description for the annotation.
        """
        original_attentions = df.bag_attn.values
        if rescale:
            df = df.copy()
            attentions = df.bag_attn.values
            df.bag_attn = (attentions - attentions.min()) / np.ptp(attentions)
        colormap = plt.get_cmap(colormap_name)
        rectangles = []
        for i, (_, row) in enumerate(df.iterrows()):
            rgba_uchar = colormap(row.bag_attn, bytes=True)
            fill_color = Color(*rgba_uchar)
            line_color = TRANSPARENT
            rectangle = Rectangle(
                f"{original_attentions[i]:.4f}",
                fill_color,
                line_color,
                left=row.left,
                top=row.top,
                right=row.right,
                bottom=row.bottom,
            )
            rectangles.append(rectangle)
        return Annotation(name, rectangles, description=description)

    @staticmethod
    def _read_csv(csv_path: Path) -> pd.DataFrame:
        return pd.read_csv(csv_path, index_col=0)

    def get_slide_annotation_from_df(
        self,
        df_or_path: Union[pd.DataFrame, Path],
        slide_id: str,
        annotation_name: str,
        **annotation_kwargs: Any,
    ) -> Annotation:
        """Create an annotation from an outputs data frame.

        :param df_or_path: Data frame or path to a CSV file.
        :param slide_id: Slide ID as described in the data frame.
        :param annotation_name: Name of the generated annotation.
        :param annotation_kwargs: Additional kwargs to :meth:`get_annotation_from_slide_data_frame`.
        """
        df = df_or_path if isinstance(df_or_path, pd.DataFrame) else self._read_csv(df_or_path)
        slide_mask = df[ResultsKey.SLIDE_ID] == slide_id
        df_slide = df[slide_mask]
        return self.get_annotation_from_slide_data_frame(df_slide, annotation_name, **annotation_kwargs)

    def upload(
        self,
        dsa: DigitalSlideArchive,
        dry_run: bool = False,
        max_slides: Optional[int] = None,
        id_filter: Optional[str] = "",
        search_mode: str = "full",
        folder_path: str = "",
        **annotation_kwargs: Any,
    ) -> List[Dict]:
        """Create annotations from a data frame and upload them to DSA.

        :param dsa: Instance of :class:`DigitalSlideArchive` to which results will be uploaded.
        :param dry_run: Create annotations and show outputs, without uploading any data.
        :param max_slides: Maximum number of slides to upload, useful for debugging.
        :param id_filter: Filter to only process slides matching this string, according to ``search_mode``.
        :param search_mode: See :meth:`DigitalSlideArchive.search_item`.
        :param folder_path: The path of the folder in DSA where results should be uploaded to.
        :param annotation_kwargs: Additional kwargs to :meth:`get_annotation_from_slide_data_frame`.
        """
        unique_slide_ids = sorted(self.df[ResultsKey.SLIDE_ID].unique())

        num_slides = len(unique_slide_ids)
        logging.info("Found outputs for %s slides", num_slides)
        if max_slides is not None:
            max_slides = min(max_slides, num_slides)
            percentage = 100 * max_slides / num_slides
            logging.info("Using %s/%s slides (%s %% of total)", max_slides, num_slides, f"{percentage:.1f}")
            unique_slide_ids = unique_slide_ids[:max_slides]

        # I think "full" is more descriptive than "text" for our API
        search_mode = "text" if search_mode == "full" else search_mode
        progress = tqdm(unique_slide_ids)
        annotation_name = f"{self.run.id} ({self.run.display_name})"
        responses = []
        if folder_path:
            folder_id = dsa.get_folder_id(folder_path)
            items = dsa.get_items_in_folder(folder_id)
            print(f"Found a total of {len(items)} items in folder {folder_path}")
        else:
            items = []
        for slide_id in progress:
            progress.set_description(slide_id)
            if id_filter not in slide_id:
                continue
            if folder_path:
                matching_items = [item for item in items if slide_id in item.name]
                if not matching_items:
                    print(f"No items in DSA for slide ID {slide_id}")
                    continue
                elif len(matching_items) > 1:
                    print(f"Multiple items in DSA for slide ID {slide_id}. Skipping this slide.")
                    continue
                item = matching_items[0]
            else:
                item = dsa.search_item(slide_id, search_mode=search_mode)

            tqdm.write(f"Processing slide {slide_id} - {item.url}")
            annotation = self.get_slide_annotation_from_df(
                self.df,
                slide_id,
                annotation_name,
                **annotation_kwargs,
            )
            if not dry_run:
                responses.append(item.add_annotation(annotation))
        return responses


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    dsa_class = DigitalSlideArchive
    parser.add_argument(
        "--run-id",
        type=str,
        help="Azure ML run ID",
    )
    parser.add_argument(
        "--dsa-url",
        type=str,
        help=f"URL to a Digital Slide Archive deployment. If not passed, it must be set in {dsa_class.URL_ENV_NAME}",
    )
    parser.add_argument(
        "--dsa-key",
        type=str,
        help=f"API key with edit permissions. If not passed, it must be set in {dsa_class.API_KEY_ENV_NAME}",
    )
    parser.add_argument(
        "--workspace-config",
        type=Path,
        help="Path to workspace configuration file (YAML or JSON)",
    )
    parser.add_argument(
        "--overwrite_csv",
        action="store_true",
        help="Force the download of the outputs CSV even when it already exists in the cache",
    )
    parser.add_argument(
        "--dry-run",
        action="store_true",
        help="Iterate over slides without uploading annotations",
    )
    parser.add_argument(
        "--max-slides",
        type=int,
        help="Maximum number of slides to process (useful for development)",
    )
    parser.add_argument(
        "--id-filter",
        type=str,
        help="Only process slides whose ID contain this substring",
        default="",
    )
    parser.add_argument(
        "--login-only",
        action="store_true",
        help="Just log into the DSA and exit. Useful to ensure connection to the DSA from current host",
    )
    parser.add_argument(
        "--no-rescale",
        action="store_false",
        default=True,
        help="Do not rescale attention values. By default, attention values are scaled such that the range "
        "between min and max fills the colormap.",
    )
    parser.add_argument(
        "--colormap",
        type=str,
        choices=plt.colormaps(),
        default="Spectral_r",
        help="Matplotlib colormap used for the heatmaps",
    )
    parser.add_argument(
        "--folder",
        type=str,
        default="",
        help="The folder in DSA where uploads should go to. Use this if there are multiple slides with the same ID"
        "in DSA. The folder name must contain the collection, like `Collection1/foo`",
    )
    parser.add_argument(
        "--split",
        type=str,
        choices=[TEST_OUTPUTS_SUBDIR, EXTRA_VAL_OUTPUTS_SUBDIR, VAL_OUTPUTS_SUBDIR],
        default=TEST_OUTPUTS_SUBDIR,
        help="The results subfolder in the AzureML run where the results are downloaded from. Default: 'test'",
    )
    parser.add_argument(
        "--search-mode",
        type=str,
        choices=("full", "prefix"),
        default="full",
        help=(
            "If \"full\", the slide ID must match the DSA file name with or without extension."
            " If \"prefix\", all files whose name starts with the slide ID will be matched"
        ),
    )
    args = parser.parse_args()

    logging_to_stdout()
    dsa = DigitalSlideArchive(args.dsa_url, args.dsa_key)
    if args.login_only:
        sys.exit(0)
    if args.run_id is None:
        raise ValueError("Please specify an Azure Machine Learning run ID")
    outputs = RunOutputs(
        run_id=args.run_id,
        workspace_config_path=args.workspace_config,
        overwrite_csv=args.overwrite_csv,
        split=args.split,
    )
    outputs.upload(
        dsa,
        dry_run=args.dry_run,
        max_slides=args.max_slides,
        id_filter=args.id_filter,
        search_mode=args.search_mode,
        colormap_name=args.colormap,
        rescale=args.no_rescale,
        folder_path=args.folder,
    )



from azure.storage.blob import generate_blob_sas, BlobSasPermissions
from azureml.core import Workspace
from datetime import datetime, timedelta
from health_azure import get_workspace
from health_ml.utils.common_utils import DEFAULT_AML_CHECKPOINT_DIR
from pathlib import Path
from typing import Optional


def get_checkpoint_url_from_aml_run(
    run_id: str,
    checkpoint_filename: str,
    account_key: str,
    expiry_days: int = 1,
    workspace_config_path: Optional[Path] = None,
    aml_workspace: Optional[Workspace] = None,
    sas_token: Optional[str] = None,
) -> str:
    """Generate a SAS URL for the checkpoint file in the given run.

    :param run_id: The run ID of the checkpoint.
    :param checkpoint_filename: The filename of the checkpoint.
    :param account_key: The Azure Storage account key to use for the SAS token.
    :param expiry_days: The number of days the SAS URL is valid for, defaults to 30.
    :param workspace_config_path: The path to the workspace config file, defaults to None.
    :param aml_workspace: The Azure ML workspace to use, defaults to the default workspace.
    :param sas_token: The SAS token to use, defaults to None.
    :return: The SAS URL for the checkpoint.
    """
    workspace = get_workspace(aml_workspace=aml_workspace, workspace_config_path=workspace_config_path)
    account_name = workspace.get_details()['storageAccount'].split('/')[-1]
    print(f"Workspace {workspace.name} stores its run results in storage account {account_name}.")
    container_name = 'azureml'
    blob_name = f'ExperimentRun/dcid.{run_id}/{DEFAULT_AML_CHECKPOINT_DIR}/{checkpoint_filename}'
    if not sas_token:
        sas_token = generate_blob_sas(
            account_name=account_name,
            container_name=container_name,
            blob_name=blob_name,
            account_key=account_key,
            permission=BlobSasPermissions(read=True),
            expiry=datetime.utcnow() + timedelta(days=expiry_days),
        )

    return f'https://{account_name}.blob.core.windows.net/{container_name}/{blob_name}?{sas_token}'


if __name__ == '__main__':
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument('--run_id', type=str, help='The run id of the model checkpoint')
    parser.add_argument('--workspace_config', type=str, help='The path to the workspace config file.')
    parser.add_argument(
        '--checkpoint_filename',
        type=str,
        default='last.ckpt',
        help='The filename of the model checkpoint. Default: last.ckpt',
    )
    parser.add_argument(
        '--expiry_days',
        type=int,
        default=30,
        help='The number of hours for which the SAS token is valid. Default: 30 for 1 month',
    )
    parser.add_argument(
        '--account_key', default='', type=str, help='The Azure Storage account key to use for the SAS token.'
    )
    args = parser.parse_args()
    workspace_config_path = Path(args.workspace_config) if args.workspace_config else None
    url = get_checkpoint_url_from_aml_run(
        run_id=args.run_id,
        checkpoint_filename=args.checkpoint_filename,
        expiry_days=args.expiry_days,
        workspace_config_path=workspace_config_path,
        account_key=args.account_key,
    )
    print(f'Checkpoint URL: {url}')



import torch
import param
from torch import Tensor
from typing import Any, Callable, List, Optional
from health_cpath.models.transforms import ExtractCoordinatesd
from health_cpath.utils.naming import ModelKey, SlideKey
from health_ml.utils.bag_utils import multibag_collate
from monai.transforms import RandGridPatchd, GridPatchd, SplitDimd


def image_collate(batch: List) -> Any:
    """
    Combine instances from a list of dicts into a single dict, by stacking them along first dim
    [{'image' : 3xHxW}, {'image' : 3xHxW}, {'image' : 3xHxW}...] - > {'image' : Nx3xHxW}
    followed by the default collate which will form a batch BxNx3xHxW.
    The list of dicts refers to the the list of tiles produced by the Rand/GridPatchd transform applied on a WSI.
    """

    for i, item in enumerate(batch):
        # The tiles have been splited into a list of dicts, each dict containing a single tile to be able to apply
        # tile wise transforms. We need to stack them back together.
        data = item[0]
        assert isinstance(data[SlideKey.IMAGE], Tensor), f"Expected torch.Tensor, got {type(data[SlideKey.IMAGE])}"
        data[SlideKey.IMAGE] = torch.stack([ix[SlideKey.IMAGE] for ix in item], dim=0)
        batch[i] = data
    return multibag_collate(batch)


class TilingParams(param.Parameterized):
    """Parameters for Tiling On the Fly a WSI using RandGridPatchd and GridPatchd monai transforms"""

    tile_size: int = param.Integer(default=224, bounds=(1, None), doc="The size of the tile, Default: 224")
    tile_overlap: int = param.Number(
        default=0,
        bounds=(0.0, 1.0),
        doc="The amount of overlap of neighboring patches in each dimension (a value between 0.0 and 1.0).",
    )
    tile_sort_fn: Optional[str] = param.String(
        default='min',
        doc="When bag_size is fixed, it determines whether to keep tiles with highest intensity values (`'max'`), "
        "lowest values (`'min'`) that assumes background is high values, or in their default order (`None`). ",
    )
    tile_pad_mode: Optional[str] = param.String(
        default=None,
        doc="The mode of padding, refer to NumpyPadMode and PytorchPadMode. Defaults to None, for no padding.",
    )
    intensity_threshold: float = param.Number(
        default=255.0, doc="The intensity threshold to filter out tiles based on intensity values. Default to None."
    )
    intensity_threshold_scale: float = param.Number(
        default=0.999, doc="The scale factor to be applied to the intensity threshold. Default to 0.999"
    )
    background_val: int = param.Integer(default=255, doc="The intensity value of background. Default to 255.")
    rand_min_offset: int = param.Integer(
        default=0,
        bounds=(0, None),
        doc="The minimum range of sarting position to be selected randomly. This parameter is passed to RandGridPatchd."
        "the random version of RandGridPatchd used at training time. Default to 0.",
    )
    rand_max_offset: int = param.Integer(
        default=None,
        bounds=(0, None),
        doc="The maximum range of sarting position to be selected randomly. This parameter is passed to RandGridPatchd."
        "the random version of RandGridPatchd used at training time. Default to None.",
    )
    inf_offset: Optional[int] = param.Integer(
        default=None,
        doc="The offset to be used for inference sampling. This parameter is passed to GridPatchd. Default to None.",
    )

    @property
    def scaled_threshold(self) -> float:
        """Returns the threshold to be used for filtering out tiles based on intensity values. We need to multiply
        the threshold by the tile size to account for the fact that the intensity is computed on the entire tile"""
        return self.intensity_threshold_scale * 3 * self.intensity_threshold * self.tile_size * self.tile_size

    def get_tiling_transform(
        self,
        bag_size: int,
        stage: ModelKey,
    ) -> Callable:
        if stage == ModelKey.TRAIN:
            return RandGridPatchd(
                keys=[SlideKey.IMAGE],
                patch_size=(self.tile_size, self.tile_size),
                min_offset=self.rand_min_offset,
                max_offset=self.rand_max_offset,
                num_patches=bag_size,
                overlap=self.tile_overlap,
                sort_fn=self.tile_sort_fn,
                threshold=self.scaled_threshold,
                pad_mode=self.tile_pad_mode,  # type: ignore
                constant_values=self.background_val,  # this arg is passed to np.pad or torch.pad
            )
        else:
            return GridPatchd(
                keys=[SlideKey.IMAGE],
                patch_size=(self.tile_size, self.tile_size),
                offset=self.inf_offset,  # type: ignore
                num_patches=bag_size,
                overlap=self.tile_overlap,
                sort_fn=self.tile_sort_fn,
                threshold=self.scaled_threshold,
                pad_mode=self.tile_pad_mode,  # type: ignore
                constant_values=self.background_val,  # this arg is passed to np.pad or torch.pad
            )

    def get_split_transform(self) -> Callable:
        """GridPatchd returns stacked tiles (bag_size, C, H, W), however we need to split them into separate
        tiles to be able to apply augmentations on each tile independently.
        """
        return SplitDimd(keys=SlideKey.IMAGE, dim=0, keepdim=False, list_output=True)

    def get_extract_coordinates_transform(self) -> Callable:
        """Extract the coordinates of the tiles returned as meta data by monai transforms to hi-ml-cpath format where
        the coordinates are represented as TileKey.TILE_LEFT, TileKey.TILE_TOP, TileKey.TILE_RIGHT, TileKey.TILE_BOTTOM.
        """
        return ExtractCoordinatesd(image_key=SlideKey.IMAGE, tile_size=self.tile_size)



#  ------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License (MIT). See LICENSE in the repo root for license information.
#  ------------------------------------------------------------------------------------------

from typing import List
import numpy as np


def location_selected_tiles(
    tile_coords: np.ndarray, location_bbox: List[int], scale_factor: int = 1, should_upscale_coords: bool = True
) -> np.ndarray:
    """Return the scaled and shifted tile co-ordinates for selected tiles in the slide.

    :param tile_coords: XY tile coordinates, assumed to be spaced by multiples of `tile_size`
    (shape: [N, 2]) in original resolution.
    :param location_bbox: Location of the bounding box on the slide in original resolution.
    :param scale_factor: Scale factor to be applied to the tile coordinates.
    :param should_upscale_coords: If True, the tile coordinates are upscaled by the scale factor before substructing
        the offset. If False, the tile coordinates are not upscaled and the offset is upscaled by the scale factor.
    """

    y_tr, x_tr = location_bbox
    if should_upscale_coords:
        tile_coords = tile_coords * scale_factor
    tile_xs, tile_ys = tile_coords.T
    tile_xs = tile_xs - x_tr
    tile_ys = tile_ys - y_tr
    tile_xs = tile_xs // scale_factor
    tile_ys = tile_ys // scale_factor

    sel_coords = np.transpose([tile_xs.tolist(), tile_ys.tolist()])

    return sel_coords



#  ------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License (MIT). See LICENSE in the repo root for license information.
#  ------------------------------------------------------------------------------------------

import logging
import sys
import math
import cv2
import numpy as np
import seaborn as sns
import matplotlib.pyplot as plt
import matplotlib.patches as patches
import matplotlib.collections as collection

from math import ceil
from pathlib import Path
from typing import Sequence, List, Any, Dict, Optional, Union, Tuple

from monai.data.dataset import Dataset
from torch.utils.data import DataLoader
from health_cpath.datasets.panda_dataset import PandaDataset
from health_cpath.preprocessing.loading import LoadingParams, ROIType, WSIBackend
from health_cpath.utils.naming import SlideKey
from health_cpath.utils.naming import ResultsKey
from health_cpath.utils.heatmap_utils import location_selected_tiles
from health_cpath.utils.tiles_selection_utils import SlideNode


logger = logging.getLogger(__name__)


def load_image_dict(sample: dict, loading_params: LoadingParams) -> Dict[SlideKey, Any]:
    """
    Load image from metadata dictionary

    :param sample: dict describing image metadata. Example:
        {'image_id': ['1ca999adbbc948e69783686e5b5414e4'],
        'image': ['/tmp/datasets/PANDA/train_images/1ca999adbbc948e69783686e5b5414e4.tiff'],
         'mask': ['/tmp/datasets/PANDA/train_label_masks/1ca999adbbc948e69783686e5b5414e4_mask.tiff'],
         'data_provider': ['karolinska'],
         'isup_grade': tensor([0]),
         'gleason_score': ['0+0']}
    :param loading_params: LoadingParams object that contains the parameters to load the image.
    """
    loader = loading_params.get_load_roid_transform()
    img = loader(sample)
    return img


def plot_panda_data_sample(
    panda_dir: str, nsamples: int, ncols: int, level: int, margin: int, title_key: str = "data_provider"
) -> None:
    """
    :param panda_dir: path to the dataset, it's expected a file called "train.csv" exists at the path.
        Look at the PandaDataset for more detail
    :param nsamples: number of random samples to be visualized
    :param ncols: number of columns in the figure grid. Nrows is automatically inferred
    :param level: level of resolution to be loaded
    :param margin: margin to be included
    :param title_key: metadata key in image_dict used to label each subplot
    """
    panda_dataset = Dataset(PandaDataset(root=panda_dir))[:nsamples]  # type: ignore
    loader = DataLoader(panda_dataset, batch_size=1)

    nrows = math.ceil(nsamples / ncols)
    fig, axes = plt.subplots(ncols=ncols, nrows=nrows, figsize=(9, 9))

    for dict_images, ax in zip(loader, axes.flat):
        slide_id = dict_images[SlideKey.SLIDE_ID]
        title = dict_images[SlideKey.METADATA][title_key]
        print(f">>> Slide {slide_id}")
        loading_params = LoadingParams(level=level, margin=margin, backend=WSIBackend.CUCIM, roi_type=ROIType.MASK)
        img = load_image_dict(dict_images, loading_params)
        ax.imshow(img[SlideKey.IMAGE].transpose(1, 2, 0))
        ax.set_title(title)
    fig.tight_layout()


def plot_scores_hist(
    results: Dict, prob_col: str = ResultsKey.CLASS_PROBS, gt_col: str = ResultsKey.TRUE_LABEL
) -> plt.Figure:
    """Plot scores as a histogram.

    :param results: List that contains slide_level dicts
    :param prob_col: column name that contains the scores
    :param gt_col: column name that contains the true label
    :return: matplotlib figure of the scores histogram by class
    """
    n_classes = len(results[prob_col][0])
    scores_class = []
    for j in range(n_classes):
        scores = [results[prob_col][i][j].cpu().item() for i, gt in enumerate(results[gt_col]) if gt == j]
        scores_class.append(scores)
    fig, ax = plt.subplots()
    ax.hist(scores_class, label=[str(i) for i in range(n_classes)], alpha=0.5)
    ax.set_xlabel("Predicted Score")
    ax.legend()
    return fig


def _get_histo_plot_title(case: str, slide_node: SlideNode) -> str:
    """Return the standard title for histopathology plots.

    :param case: case id e.g., TP, FN, FP, TN
    :param slide_node: SlideNode object that encapsulates the slide information
    """
    return (
        f"{case}: {slide_node.slide_id} P={slide_node.pred_prob_score:.2f} \n Predicted label: {slide_node.pred_label} "
        f"True label: {slide_node.true_label}"
    )


def plot_attention_tiles(
    case: str, slide_node: SlideNode, top: bool, num_columns: int, figsize: Tuple[int, int]
) -> Optional[plt.Figure]:
    """Plot top or bottom tiles figures along with their attention scores.

    :param case: The report case (e.g., TP, FN, ...)
    :param slide_node: The slide node for which we would like to plot attention tiles.
    :param top: Decides which tiles to plot. If True, plots top tile nodes of the slide_node. Otherwise, plots bottom
        tile nodes.
    :param num_columns: Number of columns to create the subfigures grid, defaults to 4
    :param figsize: The figure size of tiles attention plots, defaults to (10, 10)
    """
    tile_nodes = slide_node.top_tiles if top else slide_node.bottom_tiles
    num_rows = int(ceil(len(tile_nodes) / num_columns))
    if num_rows == 0:
        logger.warning(
            "The number of selected top and bottom tiles is too low, plotting will be skipped."
            "Try debugging with a higher num_top_tiles and/or a higher number of batches."
        )
        return None

    fig, axs = plt.subplots(nrows=num_rows, ncols=num_columns, figsize=figsize)
    fig.suptitle(_get_histo_plot_title(case, slide_node))
    for ax, tile_node in zip(axs.flat, tile_nodes):
        ax.imshow(np.transpose(tile_node.data.numpy(), (1, 2, 0)), clim=(0, 255), cmap="gray")
        ax.set_title("%.6f" % tile_node.attn)

    for ax in axs.flat:
        ax.set_axis_off()
    return fig


def plot_slide(case: str, slide_node: SlideNode, slide_image: np.ndarray, scale: float) -> plt.Figure:
    """Plots a slide thumbnail from a given slide image and scale.

    :param case: The report case (e.g., TP, FN, ...)
    :param slide_node: The slide node for which we would like to plot attention tiles.
    :param slide_image: Numpy array of the slide image (shape: [3, H, W]).
    :return: matplotlib figure of the slide thumbnail.
    """
    fig, ax = plt.subplots()
    slide_image = slide_image.transpose(1, 2, 0)
    ax.imshow(slide_image)
    fig.suptitle(_get_histo_plot_title(case, slide_node))
    ax.set_axis_off()
    original_size = fig.get_size_inches()
    fig.set_size_inches((original_size[0] * scale, original_size[1] * scale))
    return fig


def plot_heatmap_overlay(
    case: str,
    slide_node: SlideNode,
    slide_dict: Dict[SlideKey, Any],
    results: Dict[ResultsKey, List[Any]],
    tile_size: int = 224,
    should_upscale_coords: bool = True,
    extra_slide_dict: Optional[Dict[SlideKey, Any]] = None,
) -> plt.Figure:
    """Plots heatmap of selected tiles (e.g. tiles in a bag) overlay on the corresponding slide.

    :param case: The report case (e.g., TP, FN, ...)
    :param slide_node: The slide node that encapsulates the slide metadata.
    :param slide_dict: Dictionary of the slide image (shape: [3, H, W]) containing the slide image and metadata: the
        location of the bouding box of the slide and the scale factor to convert the heatmap coordinates to the slide
        level.
    :param results: Dict containing ResultsKey keys (e.g. slide id) and values as lists of output slides.
    :param tile_size: Size of each tile. Default 224.
    :param should_upscale_coords: If True, upscales the heatmap coordinates to the slide level. Default True.
    :param extra_slide_dict: An optional dictionary containing an extra slide image and metadata. Default None.
    :return: matplotlib figure of the heatmap of the given tiles on slide.
    """
    fig = plt.figure(constrained_layout=True)
    gs = fig.add_gridspec(2 if not extra_slide_dict else 3, 1)
    ax0 = fig.add_subplot(gs[0, 0])
    ax1 = fig.add_subplot(gs[1, 0], sharex=ax0, sharey=ax0)
    cax = ax1.inset_axes([1.02, 0, 0.03, 1], transform=ax1.transAxes)  # add colorbar axis
    axes = [ax0, ax1]
    if extra_slide_dict:
        ax2 = fig.add_subplot(gs[2, 0], sharex=ax0, sharey=ax0)
        axes.append(ax2)
    fig.suptitle(_get_histo_plot_title(case, slide_node))

    def _get_slide_image_from_slide_dict(slide_dict: Dict[SlideKey, Any]) -> np.ndarray:
        slide_image = slide_dict[SlideKey.IMAGE]
        assert isinstance(slide_image, np.ndarray), f"slide image must be a numpy array, got {type(slide_image)}"
        return slide_image.transpose(1, 2, 0)

    slide_image = _get_slide_image_from_slide_dict(slide_dict)

    ax0.imshow(slide_image)
    ax1.imshow(slide_image, alpha=0.5)
    if extra_slide_dict:
        extra_image = _get_slide_image_from_slide_dict(extra_slide_dict)
        if extra_image.shape != slide_image.shape:
            extra_image = cv2.resize(
                extra_image, (slide_image.shape[1], slide_image.shape[0]), interpolation=cv2.INTER_AREA
            )
        ax2.imshow(extra_image)
        ax2.tick_params('x', labelbottom=False)
    for ax in axes:
        ax.set_xlim(0, slide_image.shape[1])
        ax.set_ylim(slide_image.shape[0], 0)
    ax0.tick_params('x', labelbottom=False)
    slide_ids = [item[0] for item in results[ResultsKey.SLIDE_ID]]
    slide_idx = slide_ids.index(slide_node.slide_id)
    attentions = results[ResultsKey.BAG_ATTN][slide_idx]

    coords = np.transpose(
        [results[ResultsKey.TILE_LEFT][slide_idx].cpu().numpy(), results[ResultsKey.TILE_TOP][slide_idx].cpu().numpy()]
    )
    attentions = np.array(attentions.cpu()).reshape(-1)

    sel_coords = location_selected_tiles(
        tile_coords=coords,
        location_bbox=slide_dict[SlideKey.ORIGIN],
        scale_factor=slide_dict[SlideKey.SCALE],
        should_upscale_coords=should_upscale_coords,
    )
    cmap = plt.cm.get_cmap("Spectral_r")  # _r reverse the color map so that the highest attention is red

    tile_xs, tile_ys = sel_coords.T
    rects = [patches.Rectangle(xy, tile_size, tile_size) for xy in zip(tile_xs, tile_ys)]

    # line width is set to 0 to avoid the black border around the tiles as the tiles are already colored
    pc = collection.PatchCollection(rects, match_original=True, cmap=cmap, alpha=0.5, linewidth=0)
    pc.set_array(np.array(attentions))
    ax1.add_collection(pc)
    cb = plt.colorbar(pc, cax=cax)  # add colorbar to the right of the plot (cax)
    mean_loc = attentions.mean()  # add a horizontal line at the mean attention value
    cb.ax.hlines(mean_loc, attentions.min(), attentions.max(), color="k", linewidth=2)
    return fig


def plot_attention_histogram(case: str, slide_node: SlideNode, results: Dict[ResultsKey, List[Any]]) -> plt.Figure:
    """Plots a histogram of the attention values of the tiles in a bag.

    :param case: The report case (e.g., TP, FN, ...)
    :param slide_node: The slide node that encapsulates the slide metadata.
    :param results: Dict containing ResultsKey keys (e.g. slide id) and values as lists of output slides.
    :return: matplotlib figure of the histogram of the attention values of the tiles in a bag.
    """
    slide_ids = [item[0] for item in results[ResultsKey.SLIDE_ID]]
    slide_idx = slide_ids.index(slide_node.slide_id)
    attentions = results[ResultsKey.BAG_ATTN][slide_idx]
    fig, ax = plt.subplots()
    ax = sns.distplot(attentions.cpu().numpy().reshape(-1), kde=False)
    ax.set_xlabel("Attention scores")
    fig.suptitle(_get_histo_plot_title(case, slide_node))
    return fig


def plot_normalized_confusion_matrix(cm: np.ndarray, class_names: Sequence[str]) -> plt.Figure:
    """Plots a normalized confusion matrix and returns the figure.

    :param cm: Normalized confusion matrix to be plotted.
    :param class_names: List of class names.
    """
    fig, ax = plt.subplots()
    ax = sns.heatmap(cm, annot=True, cmap="Blues", fmt=".2%")
    ax.set_xlabel("Predicted")
    ax.set_ylabel("True")
    ax.xaxis.set_ticklabels(class_names)
    ax.yaxis.set_ticklabels(class_names)
    return fig


def plot_normalized_and_non_normalized_confusion_matrices(
    cm: np.ndarray,
    cm_n: np.ndarray,
    class_names: Sequence[str],
) -> plt.Figure:
    """Plots a normalized and non-normalized confusion matrix and returns the figure.

    :param cm: Non normalized confusion matrix to be plotted.
    :param cm_n: Normalized confusion matrix to be plotted.
    :param class_names: List of class names.
    """
    fig, axes = plt.subplots(1, 2, figsize=(15, 5))
    axes[0] = sns.heatmap(cm, annot=True, cmap="Blues", fmt="d", ax=axes[0])
    axes[1] = sns.heatmap(cm_n, annot=True, cmap="Blues", fmt=".2%", ax=axes[1])
    for ax in axes:
        ax.set_xlabel("Predicted")
        ax.set_ylabel("True")
        ax.xaxis.set_ticklabels(class_names)
        ax.yaxis.set_ticklabels(class_names)
    return fig


def resize_and_save(width_inch: int, height_inch: int, filename: Union[Path, str], dpi: int = 150) -> None:
    """
    Resizes the present figure to the given (width, height) in inches, and saves it to the given filename.

    :param width_inch: The width of the figure in inches.
    :param height_inch: The height of the figure in inches.
    :param filename: The filename to save to.
    :param dpi: Image resolution in dots per inch
    """
    fig = plt.gcf()
    fig.set_size_inches(width_inch, height_inch)
    # Workaround for Exception in Tkinter callback
    fig.canvas.start_event_loop(sys.float_info.min)
    plt.savefig(filename, dpi=dpi, bbox_inches="tight", pad_inches=0.1)


def save_figure(fig: Optional[plt.figure], figpath: Path) -> None:
    """Save a matplotlib figure in a given figpath.

    :param fig: The figure to be saved.
    :param figpath: The filename where to save the figure.
    """
    if fig is None:
        return
    fig.savefig(figpath, bbox_inches="tight")
    plt.close(fig)



#  ------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License (MIT). See LICENSE in the repo root for license information.
#  ------------------------------------------------------------------------------------------

from typing import Tuple

from torch import as_tensor, device, nn, no_grad, prod, rand
from torch.hub import load_state_dict_from_url
from torchvision.transforms import Normalize


def get_imagenet_preprocessing() -> nn.Module:
    return Normalize(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225])


def setup_feature_extractor(pretrained_model: nn.Module, input_dim: Tuple[int, ...]) -> Tuple[nn.Module, int]:
    try:
        # Attempt to auto-detect final classification layer:
        num_features: int = pretrained_model.fc.in_features  # type: ignore
        pretrained_model.fc = nn.Flatten()
        feature_extractor = pretrained_model
    except AttributeError:
        # Otherwise fallback to sequence of child modules:
        # Remove the last fully connected layer
        layers = list(pretrained_model.children())[:-1]
        layers.append(nn.Flatten())  # flatten non-batch dims in case of spatial feature maps
        feature_extractor = nn.Sequential(*layers)
        with no_grad():
            feature_shape = feature_extractor(rand(1, *input_dim)).shape
        num_features = int(prod(as_tensor(feature_shape)).item())
    # fix weights, no fine-tuning
    for param in feature_extractor.parameters():
        param.requires_grad = False
    return feature_extractor, num_features


def load_weights_to_model(weights_url: str, model: nn.Module) -> nn.Module:
    """
    Load weights to the histoSSL model from the given URL
    https://github.com/ozanciga/self-supervised-histopathology
    """
    map_location = device('cpu')
    state = load_state_dict_from_url(weights_url, map_location=map_location)  # type: ignore
    state_dict = state['state_dict']
    model_dict = model.state_dict()

    new_weights = {}
    for key, value in state_dict.items():
        model_key = key.replace('model.', '').replace('resnet.', '')
        if model_key in model_dict:
            new_weights[model_key] = value
    if len(new_weights) == 0:
        raise RuntimeError("Weights could not be loaded.")
    model_dict.update(new_weights)  # type: ignore

    model.load_state_dict(model_dict)  # type: ignore
    return model
